"""Configurable parameters for STFPM anomaly base task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from attr import attrs

from otx.algorithms.anomaly.configs.base.configuration import BaseAnomalyConfig
from otx.algorithms.anomaly.configs.base.configuration_enums import (
    EarlyStoppingMetrics,
    ModelBackbone,
)
from otx.api.configuration.elements import (
    ParameterGroup,
    add_parameter_group,
    configurable_float,
    configurable_integer,
    selectable,
    string_attribute,
)
from otx.api.configuration.model_lifecycle import ModelLifecycle


@attrs
class STFPMAnomalyBaseConfig(BaseAnomalyConfig):
    """Configurable parameters for STFPM anomaly base task."""

    header = string_attribute("Configuration for STFPM")
    description = header

    @attrs
    class LearningParameters(BaseAnomalyConfig.LearningParameters):
        """Parameters that can be tuned using HPO."""

        lr = configurable_float(
            default_value=0.4,
            header="Learning Rate",
            min_value=1e-3,
            max_value=1,
            description="Learning rate used for optimizing the Student network.",
        )

        momentum = configurable_float(
            default_value=0.9,
            header="Momentum",
            min_value=0.1,
            max_value=1.0,
            description="Momentum used for SGD optimizer",
        )

        weight_decay = configurable_float(
            default_value=0.0001,
            header="Weight Decay",
            min_value=1e-5,
            max_value=1,
            description="Decay for SGD optimizer",
        )

        backbone = selectable(
            default_value=ModelBackbone.RESNET18,
            header="Model Backbone",
            description="Pre-trained backbone used for feature extraction",
        )

        @attrs
        class EarlyStoppingParameters(ParameterGroup):
            """Early stopping parameters."""

            header = string_attribute("Early Stopping Parameters")
            description = header

            metric = selectable(
                default_value=EarlyStoppingMetrics.IMAGE_F1,
                header="Early Stopping Metric",
                description="The metric used to determine if the model should stop training",
            )

            patience = configurable_integer(
                default_value=10,
                min_value=1,
                max_value=100,
                header="Early Stopping Patience",
                description="Number of epochs to wait for an improvement in the monitored metric. If the metric has "
                "not improved for this many epochs, the training will stop and the best model will be "
                "returned.",
                warning="Setting this value too low might lead to underfitting. Setting the value too high will "
                "increase the training time and might lead to overfitting.",
                affects_outcome_of=ModelLifecycle.TRAINING,
            )

        early_stopping = add_parameter_group(EarlyStoppingParameters)

        max_epochs = configurable_integer(
            default_value=100,
            header="Max Epochs",
            min_value=1,
            max_value=500,
            description="Maximum number of epochs to train the model for.",
            warning="Training for very few epochs might lead to poor performance. If Early Stopping is enabled then "
            "increasing the value of max epochs might not lead to desired result.",
            affects_outcome_of=ModelLifecycle.TRAINING,
        )

    learning_parameters = add_parameter_group(LearningParameters)



"""Openvino Task of Segmentation."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import io
import json
import os
import tempfile
from typing import Any, Dict, Optional, Tuple, Union
from zipfile import ZipFile

import attr
import numpy as np
from addict import Dict as ADDict
from compression.api import DataLoader
from compression.engines.ie_engine import IEEngine
from compression.graph import load_model, save_model
from compression.graph.model_utils import compress_model_weights, get_nodes_by_type
from compression.pipeline.initializer import create_pipeline
from openvino.model_zoo.model_api.adapters import OpenvinoAdapter, create_core
from openvino.model_zoo.model_api.models import Model

from otx.algorithms.segmentation.adapters.openvino import model_wrappers
from otx.algorithms.segmentation.adapters.openvino.model_wrappers.blur import (
    get_activation_map,
)
from otx.algorithms.segmentation.configs.base import SegmentationConfig
from otx.api.entities.annotation import AnnotationSceneEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import (
    InferenceParameters,
    default_progress_callback,
)
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
    OptimizationMethod,
)
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.serialization.label_mapper import LabelSchemaMapper, label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.exportable_code import demo
from otx.api.usecases.exportable_code.inference import BaseInferencer
from otx.api.usecases.exportable_code.prediction_to_annotation_converter import (
    SegmentationToAnnotationConverter,
)
from otx.api.usecases.tasks.interfaces.deployment_interface import IDeploymentTask
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.mpa.utils.logger import get_logger

logger = get_logger()


# pylint: disable=too-many-locals, too-many-statements, unused-argument
class OpenVINOSegmentationInferencer(BaseInferencer):
    """Inferencer implementation for Segmentation using OpenVINO backend."""

    @check_input_parameters_type()
    def __init__(
        self,
        hparams: SegmentationConfig,
        label_schema: LabelSchemaEntity,
        model_file: Union[str, bytes],
        weight_file: Union[str, bytes, None] = None,
        device: str = "CPU",
        num_requests: int = 1,
    ):
        """Inferencer implementation for Segmentation using OpenVINO backend.

        :param hparams: Hyper parameters that the model should use.
        :param label_schema: LabelSchemaEntity that was used during model training.
        :param model_file: Path to model to load, `.xml`, `.bin` or `.onnx` file.
        :param device: Device to run inference on, such as CPU, GPU or MYRIAD. Defaults to "CPU".
        :param num_requests: Maximum number of requests that the inferencer can make.
            Good value is the number of available cores. Defaults to 1.
        """

        model_adapter = OpenvinoAdapter(
            create_core(), model_file, weight_file, device=device, max_num_requests=num_requests
        )
        self.configuration = {
            **attr.asdict(
                hparams.postprocessing,
                filter=lambda attr, value: attr.name
                not in ["header", "description", "type", "visible_in_ui", "class_name"],
            )
        }
        self.model = Model.create_model(
            hparams.postprocessing.class_name.value, model_adapter, self.configuration, preload=True
        )
        self.converter = SegmentationToAnnotationConverter(label_schema)

    @check_input_parameters_type()
    def pre_process(self, image: np.ndarray) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
        """Pre-process function of OpenVINO Segmentation Inferencer."""
        return self.model.preprocess(image)

    @check_input_parameters_type()
    def post_process(
        self, prediction: Dict[str, np.ndarray], metadata: Dict[str, Any]
    ) -> Tuple[AnnotationSceneEntity, Any, Any]:
        """Post-process function of OpenVINO Segmentation Inferencer."""
        hard_prediction = self.model.postprocess(prediction, metadata)
        soft_prediction = metadata["soft_prediction"]
        feature_vector = metadata["feature_vector"]
        predicted_scene = self.converter.convert_to_annotation(hard_prediction, metadata)

        return predicted_scene, feature_vector, soft_prediction

    @check_input_parameters_type()
    def predict(self, image: np.ndarray) -> Tuple[AnnotationSceneEntity, Any, Any]:
        """Perform a prediction for a given input image."""
        image, metadata = self.pre_process(image)
        predictions = self.forward(image)
        predictions = self.post_process(predictions, metadata)
        return predictions

    @check_input_parameters_type()
    def forward(self, image: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """Forward function of OpenVINO Segmentation Inferencer."""
        return self.model.infer_sync(image)


class OTXOpenVinoDataLoader(DataLoader):
    """Data loader for OTXDetection using OpenVINO backend."""

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def __init__(self, dataset: DatasetEntity, inferencer: BaseInferencer):
        self.dataset = dataset
        self.inferencer = inferencer

    @check_input_parameters_type()
    def __getitem__(self, index: int):
        """Return dataset item from index."""
        image = self.dataset[index].numpy
        annotation = self.dataset[index].annotation_scene
        inputs, metadata = self.inferencer.pre_process(image)

        return (index, annotation), inputs, metadata

    def __len__(self):
        """Length of OTXOpenVinoDataLoader."""
        return len(self.dataset)


class OpenVINOSegmentationTask(IDeploymentTask, IInferenceTask, IEvaluationTask, IOptimizationTask):
    """Task implementation for Segmentation using OpenVINO backend."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment):
        self.task_environment = task_environment
        self.model = self.task_environment.model
        self.model_name = self.task_environment.model_template.model_template_id
        self.inferencer = self.load_inferencer()

        labels = task_environment.get_labels(include_empty=False)
        self._label_dictionary = dict(enumerate(labels, 1))
        template_file_path = self.task_environment.model_template.model_template_path
        self._base_dir = os.path.abspath(os.path.dirname(template_file_path))

    @property
    def hparams(self):
        """Hparams of OpenVINO Segmentation Task."""
        return self.task_environment.get_hyper_parameters(SegmentationConfig)

    def load_inferencer(self) -> OpenVINOSegmentationInferencer:
        """load_inferencer function of OpenVINO Segmentation Task."""
        if self.model is None:
            raise RuntimeError("load_inferencer failed, model is None")
        return OpenVINOSegmentationInferencer(
            self.hparams,
            self.task_environment.label_schema,
            self.model.get_data("openvino.xml"),
            self.model.get_data("openvino.bin"),
        )

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def infer(
        self, dataset: DatasetEntity, inference_parameters: Optional[InferenceParameters] = None
    ) -> DatasetEntity:
        """Infer function of OpenVINOSegmentationTask."""
        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress
            dump_soft_prediction = not inference_parameters.is_evaluation
        else:
            update_progress_callback = default_progress_callback
            dump_soft_prediction = True

        dataset_size = len(dataset)
        for i, dataset_item in enumerate(dataset, 1):
            predicted_scene, feature_vector, soft_prediction = self.inferencer.predict(dataset_item.numpy)
            dataset_item.append_annotations(predicted_scene.annotations)

            if feature_vector is not None:
                feature_vector_media = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(feature_vector_media, model=self.model)

            if dump_soft_prediction:
                for label_index, label in self._label_dictionary.items():
                    if label_index == 0:
                        continue
                    current_label_soft_prediction = soft_prediction[:, :, label_index]
                    class_act_map = get_activation_map(current_label_soft_prediction)
                    result_media = ResultMediaEntity(
                        name=label.name,
                        type="soft_prediction",
                        label=label,
                        annotation_scene=dataset_item.annotation_scene,
                        roi=dataset_item.roi,
                        numpy=class_act_map,
                    )
                    dataset_item.append_metadata_item(result_media, model=self.model)

            update_progress_callback(int(i / dataset_size * 100), None)

        return dataset

    @check_input_parameters_type()
    def evaluate(self, output_resultset: ResultSetEntity, evaluation_metric: Optional[str] = None):
        """Evaluate function of OpenVINOSegmentationTask."""
        logger.info("Computing mDice")
        metrics = MetricsHelper.compute_dice_averaged_over_pixels(output_resultset)
        logger.info(f"mDice after evaluation: {metrics.overall_dice.value}")

        output_resultset.performance = metrics.get_performance()

    @check_input_parameters_type()
    def deploy(self, output_model: ModelEntity) -> None:
        """Deploy function of OpenVINOSegmentationTask."""
        logger.info("Deploying the model")
        if self.model is None:
            raise RuntimeError("deploy failed, model is None")

        work_dir = os.path.dirname(demo.__file__)
        parameters = {}
        parameters["type_of_model"] = self.hparams.postprocessing.class_name.value
        parameters["converter_type"] = "SEGMENTATION"
        parameters["model_parameters"] = self.inferencer.configuration
        parameters["model_parameters"]["labels"] = LabelSchemaMapper.forward(self.task_environment.label_schema)

        zip_buffer = io.BytesIO()
        with ZipFile(zip_buffer, "w") as arch:
            # model files
            arch.writestr(os.path.join("model", "model.xml"), self.model.get_data("openvino.xml"))
            arch.writestr(os.path.join("model", "model.bin"), self.model.get_data("openvino.bin"))
            arch.writestr(os.path.join("model", "config.json"), json.dumps(parameters, ensure_ascii=False, indent=4))
            # model_wrappers files
            for root, _, files in os.walk(os.path.dirname(model_wrappers.__file__)):
                for file in files:
                    file_path = os.path.join(root, file)
                    arch.write(
                        file_path, os.path.join("python", "model_wrappers", file_path.split("model_wrappers/")[1])
                    )
            # other python files
            arch.write(os.path.join(work_dir, "requirements.txt"), os.path.join("python", "requirements.txt"))
            arch.write(os.path.join(work_dir, "LICENSE"), os.path.join("python", "LICENSE"))
            arch.write(os.path.join(work_dir, "README.md"), os.path.join("python", "README.md"))
            arch.write(os.path.join(work_dir, "demo.py"), os.path.join("python", "demo.py"))
        output_model.exportable_code = zip_buffer.getvalue()
        logger.info("Deploying completed")

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def optimize(
        self,
        optimization_type: OptimizationType,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        optimization_parameters: Optional[OptimizationParameters] = None,
    ):
        """Optimize function of OpenVINOSegmentationTask."""
        logger.info("Start POT optimization")
        if self.model is None:
            raise RuntimeError("POT optimize failed, model is None")

        if optimization_type is not OptimizationType.POT:
            raise ValueError("POT is the only supported optimization type for OpenVino models")

        dataset = dataset.get_subset(Subset.TRAINING)
        data_loader = OTXOpenVinoDataLoader(dataset, self.inferencer)

        with tempfile.TemporaryDirectory() as tempdir:
            xml_path = os.path.join(tempdir, "model.xml")
            bin_path = os.path.join(tempdir, "model.bin")
            with open(xml_path, "wb") as f:
                f.write(self.model.get_data("openvino.xml"))
            with open(bin_path, "wb") as f:
                f.write(self.model.get_data("openvino.bin"))

            model_config = ADDict({"model_name": "openvino_model", "model": xml_path, "weights": bin_path})

            model = load_model(model_config)

            if get_nodes_by_type(model, ["FakeQuantize"]):
                raise RuntimeError("Model is already optimized by POT")

        if optimization_parameters is not None:
            optimization_parameters.update_progress(10, None)

        engine_config = ADDict({"device": "CPU"})

        optimization_config_path = os.path.join(self._base_dir, "pot_optimization_config.json")
        if os.path.exists(optimization_config_path):
            with open(optimization_config_path, encoding="UTF-8") as f_src:
                algorithms = ADDict(json.load(f_src))["algorithms"]
        else:
            algorithms = [ADDict({"name": "DefaultQuantization", "params": {"target_device": "ANY"}})]
        for algo in algorithms:
            algo.params.stat_subset_size = self.hparams.pot_parameters.stat_subset_size
            algo.params.shuffle_data = True
            if "Quantization" in algo["name"]:
                algo.params.preset = self.hparams.pot_parameters.preset.name.lower()

        engine = IEEngine(config=engine_config, data_loader=data_loader, metric=None)

        pipeline = create_pipeline(algorithms, engine)

        compressed_model = pipeline.run(model)

        compress_model_weights(compressed_model)

        if optimization_parameters is not None:
            optimization_parameters.update_progress(90, None)

        with tempfile.TemporaryDirectory() as tempdir:
            save_model(compressed_model, tempdir, model_name="model")
            with open(os.path.join(tempdir, "model.xml"), "rb") as f:
                output_model.set_data("openvino.xml", f.read())
            with open(os.path.join(tempdir, "model.bin"), "rb") as f:
                output_model.set_data("openvino.bin", f.read())

        output_model.set_data("label_schema.json", label_schema_to_bytes(self.task_environment.label_schema))

        # set model attributes for quantized model
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.POT
        output_model.optimization_methods = [OptimizationMethod.QUANTIZATION]
        output_model.precision = [ModelPrecision.INT8]

        self.model = output_model
        self.inferencer = self.load_inferencer()

        if optimization_parameters is not None:
            optimization_parameters.update_progress(100, None)
        logger.info("POT optimization completed")



"""This module implements segmentation related utilities."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import warnings
from copy import copy
from typing import List, Optional, Sequence, Tuple, cast

import cv2
import numpy as np
from bson import ObjectId

from otx.api.entities.annotation import Annotation
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.id import ID
from otx.api.entities.label import LabelEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.utils.shape_factory import ShapeFactory


def mask_from_dataset_item(dataset_item: DatasetItemEntity, labels: List[LabelEntity]) -> np.ndarray:
    """Creates a mask from dataset item.

    The mask will be two dimensional, and the value of each pixel matches the class index with offset 1. The background
    class index is zero. labels[0] matches pixel value 1, etc. The class index is
    determined based on the order of 'labels'.

    Args:
        dataset_item: Item to make mask for
        labels: The labels to use for creating the mask. The order of
            the labels determines the class index.

    Returns:
        Numpy array of mask
    """
    # todo: cache this so that it does not have to be redone for all the same media
    mask = mask_from_annotation(
        dataset_item.get_annotations(),
        labels,
        dataset_item.width,
        dataset_item.height,
    )

    return mask


def mask_from_annotation(
    annotations: List[Annotation], labels: List[LabelEntity], width: int, height: int
) -> np.ndarray:
    """Generate a segmentation mask of a numpy image, and a list of shapes.

    The mask is will be two dimensional and the value of each pixel matches the class
    index with offset 1. The background class index is zero. labels[0] matches pixel
    value 1, etc. The class index is determined based on the order of `labels`:

    Args:
        annotations: List of annotations to plot in mask
        labels: List of labels. The index position of the label
            determines the class number in the segmentation mask.
        width: Width of the mask
        height: Height of the mask

    Returns:
        2d numpy array of mask
    """

    labels = sorted(labels)  # type: ignore
    mask = np.zeros(shape=(height, width), dtype=np.uint8)
    for annotation in annotations:
        shape = annotation.shape
        if not isinstance(shape, Polygon):
            shape = ShapeFactory.shape_as_polygon(annotation.shape)
        known_labels = [
            label for label in annotation.get_labels() if isinstance(label, ScoredLabel) and label.get_label() in labels
        ]
        if len(known_labels) == 0:
            # Skip unknown shapes
            continue

        label_to_compare = known_labels[0].get_label()

        class_idx = labels.index(label_to_compare) + 1
        contour = []
        for point in shape.points:
            contour.append([int(point.x * width), int(point.y * height)])

        mask = cv2.drawContours(mask, np.asarray([contour]), 0, (class_idx, class_idx, class_idx), -1)

    mask = np.expand_dims(mask, axis=2)

    return mask


def create_hard_prediction_from_soft_prediction(
    soft_prediction: np.ndarray, soft_threshold: float, blur_strength: int = 5
) -> np.ndarray:
    """Creates a hard prediction containing the final label index per pixel.

    Args:
        soft_prediction: Output from segmentation network. Assumes
            floating point values, between 0.0 and 1.0. Can be a
            2d-array of shape (height, width) or per-class segmentation
            logits of shape (height, width, num_classes)
        soft_threshold: minimum class confidence for each pixel. The
            higher the value, the more strict the segmentation is
            (usually set to 0.5)
        blur_strength: The higher the value, the smoother the
            segmentation output will be, but less accurate

    Returns:
        Numpy array of the hard prediction
    """
    soft_prediction_blurred = cv2.blur(soft_prediction, (blur_strength, blur_strength))
    if len(soft_prediction.shape) == 3:
        # Apply threshold to filter out `unconfident` predictions, then get max along
        # class dimension
        soft_prediction_blurred[soft_prediction_blurred < soft_threshold] = 0
        hard_prediction = np.argmax(soft_prediction_blurred, axis=2)
    elif len(soft_prediction.shape) == 2:
        # In the binary case, simply apply threshold
        hard_prediction = soft_prediction_blurred > soft_threshold
    else:
        raise ValueError(
            f"Invalid prediction input of shape {soft_prediction.shape}. " f"Expected either a 2D or 3D array."
        )
    return hard_prediction


Contour = List[Tuple[float, float]]


def get_subcontours(contour: Contour) -> List[Contour]:
    """Splits contour into subcontours that do not have self intersections."""

    ContourInternal = List[Optional[Tuple[float, float]]]

    def find_loops(points: ContourInternal) -> List[Sequence[int]]:
        """For each consecutive pair of equivalent rows in the input matrix returns their indices."""
        _, inverse, count = np.unique(points, axis=0, return_inverse=True, return_counts=True)
        duplicates = np.where(count > 1)[0]
        indices = []
        for x in duplicates:
            y = np.nonzero(inverse == x)[0]
            for i, _ in enumerate(y[:-1]):
                indices.append(y[i : i + 2])
        return indices

    base_contour = cast(ContourInternal, copy(contour))

    # Make sure that contour is closed.
    if not np.array_equal(base_contour[0], base_contour[-1]):
        base_contour.append(base_contour[0])

    subcontours: List[Contour] = []
    loops = sorted(find_loops(base_contour), key=lambda x: x[0], reverse=True)
    for loop in loops:
        i, j = loop
        subcontour = base_contour[i:j]
        subcontour = list(x for x in subcontour if x is not None)
        subcontours.append(cast(Contour, subcontour))
        base_contour[i:j] = [None] * (j - i)

    subcontours = [i for i in subcontours if len(i) > 2]
    return subcontours


def create_annotation_from_segmentation_map(
    hard_prediction: np.ndarray, soft_prediction: np.ndarray, label_map: dict
) -> List[Annotation]:
    """Creates polygons from the soft predictions.

    Background label will be ignored and not be converted to polygons.

    Args:
        hard_prediction: hard prediction containing the final label
            index per pixel. See function
            `create_hard_prediction_from_soft_prediction`.
        soft_prediction: soft prediction with shape H x W x N_labels,
            where soft_prediction[:, :, 0] is the soft prediction for
            background. If soft_prediction is of H x W shape, it is
            assumed that this soft prediction will be applied for all
            labels.
        label_map: dictionary mapping labels to an index. It is assumed
            that the first item in the dictionary corresponds to the
            background label and will therefore be ignored.

    Returns:
        List of shapes
    """
    # pylint: disable=too-many-locals
    height, width = hard_prediction.shape[:2]
    img_class = hard_prediction.swapaxes(0, 1)

    # pylint: disable=too-many-nested-blocks
    annotations: List[Annotation] = []
    for label_index, label in label_map.items():
        # Skip background
        if label_index == 0:
            continue

        # obtain current label soft prediction
        if len(soft_prediction.shape) == 3:
            current_label_soft_prediction = soft_prediction[:, :, label_index]
        else:
            current_label_soft_prediction = soft_prediction

        obj_group = img_class == label_index
        label_index_map = (obj_group.T.astype(int) * 255).astype(np.uint8)

        # Contour retrieval mode CCOMP (Connected components) creates a two-level
        # hierarchy of contours
        contours, hierarchies = cv2.findContours(label_index_map, cv2.RETR_CCOMP, cv2.CHAIN_APPROX_NONE)

        if hierarchies is not None:
            for contour, hierarchy in zip(contours, hierarchies[0]):
                if hierarchy[3] == -1:
                    # In this case a contour does not represent a hole
                    contour = list((point[0][0], point[0][1]) for point in contour)

                    # Split contour into subcontours that do not have self intersections.
                    subcontours = get_subcontours(contour)

                    for subcontour in subcontours:
                        # compute probability of the shape
                        mask = np.zeros(hard_prediction.shape, dtype=np.uint8)
                        cv2.drawContours(
                            mask,
                            np.asarray([[[x, y]] for x, y in subcontour]),
                            contourIdx=-1,
                            color=1,
                            thickness=-1,
                        )
                        probability = cv2.mean(current_label_soft_prediction, mask)[0]

                        # convert the list of points to a closed polygon
                        points = [Point(x=x / width, y=y / height) for x, y in subcontour]
                        polygon = Polygon(points=points)

                        if polygon.get_area() > 0:
                            # Contour is a closed polygon with area > 0
                            annotations.append(
                                Annotation(
                                    shape=polygon,
                                    labels=[ScoredLabel(label, probability)],
                                    id=ID(ObjectId()),
                                )
                            )
                        else:
                            # Contour is a closed polygon with area == 0
                            warnings.warn(
                                "The geometry of the segmentation map you are converting "
                                "is not fully supported. Polygons with a area of zero "
                                "will be removed.",
                                UserWarning,
                            )
                else:
                    # If contour hierarchy[3] != -1 then contour has a parent and
                    # therefore is a hole
                    # Do not allow holes in segmentation masks to be filled silently,
                    # but trigger warning instead
                    warnings.warn(
                        "The geometry of the segmentation map you are converting is "
                        "not fully supported. A hole was found and will be filled.",
                        UserWarning,
                    )

    return annotations



"""Configuration file of OTX Detection."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from attr import attrs

from otx.algorithms.common.configs import BaseConfig, LearningRateSchedule
from otx.api.configuration.elements import (
    add_parameter_group,
    boolean_attribute,
    selectable,
    string_attribute,
)

# pylint: disable=invalid-name


@attrs
class DetectionConfig(BaseConfig):
    """Configurations of OTX Detection."""

    header = string_attribute("Configuration for an object detection task of MPA")
    description = header

    @attrs
    class __LearningParameters(BaseConfig.BaseLearningParameters):
        header = string_attribute("Learning Parameters")
        description = header

        learning_rate_schedule = selectable(
            default_value=LearningRateSchedule.COSINE,
            header="Learning rate schedule",
            description="Specify learning rate scheduling for the MMDetection task. "
            "When training for a small number of epochs (N < 10), the fixed "
            "schedule is recommended. For training for 10 < N < 25 epochs, "
            "step-wise or exponential annealing might give better results. "
            "Finally, for training on large datasets for at least 20 "
            "epochs, cyclic annealing could result in the best model.",
            editable=True,
            visible_in_ui=True,
        )

    @attrs
    class __Postprocessing(BaseConfig.BasePostprocessing):
        header = string_attribute("Postprocessing")
        description = header

    @attrs
    class __NNCFOptimization(BaseConfig.BaseNNCFOptimization):
        header = string_attribute("Optimization by NNCF")
        description = header
        visible_in_ui = boolean_attribute(False)

    @attrs
    class __POTParameter(BaseConfig.BasePOTParameter):
        header = string_attribute("POT Parameters")
        description = header
        visible_in_ui = boolean_attribute(False)

    @attrs
    class __AlgoBackend(BaseConfig.BaseAlgoBackendParameters):
        header = string_attribute("Parameters for the MPA algo-backend")
        description = header

    @attrs
    class __TilingParameters(BaseConfig.BaseTilingParameters):
        header = string_attribute("Tiling Parameters")
        description = header

    learning_parameters = add_parameter_group(__LearningParameters)
    postprocessing = add_parameter_group(__Postprocessing)
    nncf_optimization = add_parameter_group(__NNCFOptimization)
    pot_parameters = add_parameter_group(__POTParameter)
    algo_backend = add_parameter_group(__AlgoBackend)
    tiling_parameters = add_parameter_group(__TilingParameters)



import os

import pytest

from otx.mpa import Stage
from otx.mpa.seg.stage import SegStage
from otx.mpa.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.segmentation.test_helpers import (
    DEFAULT_RECIPE_CONFIG_PATH,
    DEFAULT_SEG_TEMPLATE_DIR,
)


class TestOTXSegStage:
    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        cfg = MPAConfig.fromfile(DEFAULT_RECIPE_CONFIG_PATH)
        self.stage = SegStage(name="", mode="train", config=cfg, common_cfg=None, index=0)
        self.model_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "model.py"))
        self.data_cfg = MPAConfig.fromfile(os.path.join(DEFAULT_SEG_TEMPLATE_DIR, "data_pipeline.py"))

    @e2e_pytest_unit
    def test_configure(self, mocker):
        mock_cfg_model = mocker.patch.object(SegStage, "configure_model")
        mock_cfg_ckpt = mocker.patch.object(SegStage, "configure_ckpt")
        mock_cfg_data = mocker.patch.object(SegStage, "configure_data")
        mock_cfg_task = mocker.patch.object(SegStage, "configure_task")
        mock_cfg_hook = mocker.patch.object(SegStage, "configure_hook")

        fake_arg = {"pretrained": True, "foo": "bar"}
        returned_value = self.stage.configure(self.model_cfg, "", self.data_cfg, True, **fake_arg)
        mock_cfg_model.assert_called_once_with(self.stage.cfg, self.model_cfg, True, **fake_arg)
        mock_cfg_ckpt.assert_called_once_with(self.stage.cfg, "", fake_arg.get("pretrained", None))
        mock_cfg_data.assert_called_once_with(self.stage.cfg, True, self.data_cfg)
        mock_cfg_task.assert_called_once_with(self.stage.cfg, True, **fake_arg)
        mock_cfg_hook.assert_called_once_with(self.stage.cfg)

        assert returned_value == self.stage.cfg

    @e2e_pytest_unit
    def test_configure_model(self):
        fake_arg = {"ir_model_path": {"ir_weight_path": "", "ir_weight_init": ""}}
        self.stage.configure_model(self.stage.cfg, self.model_cfg, True, **fake_arg)

        assert self.stage.cfg.model_task

    @e2e_pytest_unit
    def test_configure_data(self, mocker):
        mock_super_cfg_data = mocker.patch.object(Stage, "configure_data")
        self.stage.configure_data(self.stage.cfg, True, self.data_cfg, pretrained=None)

        mock_super_cfg_data.assert_called_once()
        assert self.stage.cfg.data
        assert self.stage.cfg.data.train
        assert self.stage.cfg.data.val

    @e2e_pytest_unit
    def test_configure_task(self, mocker):
        mock_cfg_classes = mocker.patch.object(SegStage, "configure_classes")
        mock_cfg_ignore = mocker.patch.object(SegStage, "configure_ignore")
        self.stage.configure_task(self.stage.cfg, True)

        mock_cfg_classes.assert_called_once()
        mock_cfg_ignore.assert_called_once()

    @e2e_pytest_unit
    def test_configure_classes_replace(self, mocker):
        mocker.patch.object(Stage, "get_data_classes", return_value=["foo", "bar"])
        self.stage.configure_classes(self.stage.cfg, "REPLACE")

        assert "background" in self.stage.model_classes
        assert self.stage.model_classes == ["background", "foo", "bar"]

    @e2e_pytest_unit
    def test_configure_classes_merge(self, mocker):
        mocker.patch.object(Stage, "get_model_classes", return_value=["foo", "bar"])
        mocker.patch.object(Stage, "get_data_classes", return_value=["foo", "baz"])
        self.stage.configure_classes(self.stage.cfg, "MERGE")

        assert "background" in self.stage.model_classes
        assert self.stage.model_classes == ["background", "foo", "bar", "baz"]

    @e2e_pytest_unit
    def test_configure_ignore(self):
        self.stage.configure_ignore(self.stage.cfg)

        if "decode_head" in self.stage.cfg.model:
            assert self.stage.cfg.model.decode_head.loss_decode.type == "CrossEntropyLossWithIgnore"



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import importlib
import json
import os
import os.path as osp
import random
import time
from typing import Any, Callable, Dict, Optional

import mmcv
import numpy as np
import torch
from mmcv import Config, ConfigDict
from mmcv.runner import CheckpointLoader, wrap_fp16_model
from torch import distributed as dist
from torch.utils.data import Dataset

from otx.algorithms.common.adapters.mmcv.utils import (
    build_dataloader,
    build_dataset,
    get_data_cfg,
)

from .registry import STAGES
from .utils.config_utils import MPAConfig, update_or_add_custom_hook
from .utils.logger import config_logger, get_logger

logger = get_logger()


def _set_random_seed(seed, deterministic=False):
    """Set random seed.

    Args:
        seed (int): Seed to be used.
        deterministic (bool): Whether to set the deterministic option for
            CUDNN backend, i.e., set `torch.backends.cudnn.deterministic`
            to True and `torch.backends.cudnn.benchmark` to False.
            Default: False.
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    logger.info(f"Training seed was set to {seed} w/ deterministic={deterministic}.")
    if deterministic:
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def get_available_types():
    types = []
    for k, v in STAGES.module_dict.items():
        # logger.info(f'key [{k}] = value[{v}]')
        types.append(k)
    return types


MODEL_TASK = {"classification": "mmcls", "detection": "mmdet", "segmentation": "mmseg"}

# @STAGES.register_module()
class Stage(object):
    MODEL_BUILDER = None

    def __init__(self, name, mode, config, common_cfg={}, index=0, **kwargs):
        logger.debug(f"init stage with: {name}, {mode}, {config}, {common_cfg}, {index}, {kwargs}")
        # the name of 'config' cannot be changed to such as 'config_file'
        # because it is defined as 'config' in recipe file.....
        self.name = name
        self.mode = mode
        self.index = index
        self.input = kwargs.pop("input", {})  # input_map?? input_dict? just input?
        self.output_keys = kwargs.pop("output", [])
        self._distributed = False

        if common_cfg is None:
            common_cfg = dict(output_path="logs")

        if not isinstance(common_cfg, dict):
            raise TypeError(f"common_cfg should be the type of dict but {type(common_cfg)}")
        else:
            if common_cfg.get("output_path") is None:
                logger.info("output_path is not set in common_cfg. set it to 'logs' as default")
                common_cfg["output_path"] = "logs"

        self.output_prefix = common_cfg["output_path"]
        self.output_suffix = f"stage{self.index:02d}_{self.name}"

        # # Work directory
        # work_dir = os.path.join(self.output_prefix, self.output_suffix)
        # mmcv.mkdir_or_exist(os.path.abspath(work_dir))

        if isinstance(config, Config):
            cfg = config
        elif isinstance(config, dict):
            cfg = Config(cfg_dict=config)
        elif isinstance(config, str):
            if os.path.exists(config):
                cfg = MPAConfig.fromfile(config)
            else:
                err_msg = f"cannot find configuration file {config}"
                logger.error(err_msg)
                raise ValueError(err_msg)
        else:
            err_msg = "'config' argument could be one of the \
                       [dictionary, Config object, or string of the cfg file path]"
            logger.error(err_msg)
            raise ValueError(err_msg)

        cfg.merge_from_dict(common_cfg)

        if len(kwargs) > 0:
            addtional_dict = {}
            logger.info("found override configurations for the stage")
            for k, v in kwargs.items():
                addtional_dict[k] = v
                logger.info(f"\t{k}: {v}")
            cfg.merge_from_dict(addtional_dict)

        max_epochs = -1
        if hasattr(cfg, "total_epochs"):
            max_epochs = cfg.pop("total_epochs")
        if hasattr(cfg, "runner"):
            if hasattr(cfg.runner, "max_epochs"):
                if max_epochs != -1:
                    max_epochs = min(max_epochs, cfg.runner.max_epochs)
                else:
                    max_epochs = cfg.runner.max_epochs
        if max_epochs > 0:
            if cfg.runner.max_epochs != max_epochs:
                cfg.runner.max_epochs = max_epochs
                logger.info(f"The maximum number of epochs is adjusted to {max_epochs}.")
            if hasattr(cfg, "checkpoint_config"):
                if hasattr(cfg.checkpoint_config, "interval"):
                    if cfg.checkpoint_config.interval > max_epochs:
                        logger.warning(
                            f"adjusted checkpoint interval from {cfg.checkpoint_config.interval} to {max_epochs} \
                            since max_epoch is shorter than ckpt interval configuration"
                        )
                        cfg.checkpoint_config.interval = max_epochs

        if cfg.get("seed", None) is not None:
            _set_random_seed(cfg.seed, deterministic=cfg.get("deterministic", False))
        else:
            cfg.seed = None

        # Work directory
        work_dir = cfg.get("work_dir", "")
        work_dir = os.path.join(self.output_prefix, work_dir if work_dir else "", self.output_suffix)
        cfg.work_dir = os.path.abspath(work_dir)
        logger.info(f"work dir = {cfg.work_dir}")
        mmcv.mkdir_or_exist(os.path.abspath(work_dir))

        # config logger replace hook
        hook_cfg = ConfigDict(type="LoggerReplaceHook")
        update_or_add_custom_hook(cfg, hook_cfg)

        self.cfg = cfg

        self.__init_device()

    def __init_device(self):
        if torch.distributed.is_initialized():
            self._distributed = True
            self.cfg.gpu_ids = [int(os.environ["LOCAL_RANK"])]
        elif "gpu_ids" not in self.cfg:
            gpu_ids = os.environ.get("CUDA_VISIBLE_DEVICES")
            logger.info(f"CUDA_VISIBLE_DEVICES = {gpu_ids}")
            if gpu_ids is not None:
                self.cfg.gpu_ids = range(len(gpu_ids.split(",")))
            else:
                self.cfg.gpu_ids = range(1)

        # consider "cuda" and "cpu" device only
        if not torch.cuda.is_available():
            self.cfg.device = "cpu"
            self.cfg.gpu_ids = range(-1, 0)
        else:
            self.cfg.device = "cuda"

    @property
    def distributed(self):
        return self._distributed

    def run(self, **kwargs):
        raise NotImplementedError

    def _init_logger(self, **kwargs):
        timestamp = time.strftime("%Y%m%d_%H%M%S", time.localtime())
        config_logger(os.path.join(self.cfg.work_dir, f"{timestamp}.log"), level=self.cfg.log_level)
        logger.info(f"configured logger at {self.cfg.work_dir} with named {timestamp}.log")
        return logger

    @staticmethod
    def configure_data(cfg, training, **kwargs):
        # update data configuration using image options
        def configure_split(target):
            def update_transform(opt, pipeline, idx, transform):
                if isinstance(opt, dict):
                    if "_delete_" in opt.keys() and opt.get("_delete_", False):
                        # if option include _delete_=True, remove this transform from pipeline
                        logger.info(f"configure_data: {transform['type']} is deleted")
                        del pipeline[idx]
                        return
                    logger.info(f"configure_data: {transform['type']} is updated with {opt}")
                    transform.update(**opt)

            def update_config(src, pipeline_options):
                logger.info(f"update_config() {pipeline_options}")
                if src.get("pipeline") is not None or (
                    src.get("dataset") is not None and src.get("dataset").get("pipeline") is not None
                ):
                    if src.get("pipeline") is not None:
                        pipeline = src.get("pipeline", None)
                    else:
                        pipeline = src.get("dataset").get("pipeline")
                    if isinstance(pipeline, list):
                        for idx, transform in enumerate(pipeline):
                            for opt_key, opt in pipeline_options.items():
                                if transform["type"] == opt_key:
                                    update_transform(opt, pipeline, idx, transform)
                    elif isinstance(pipeline, dict):
                        for _, pipe in pipeline.items():
                            for idx, transform in enumerate(pipe):
                                for opt_key, opt in pipeline_options.items():
                                    if transform["type"] == opt_key:
                                        update_transform(opt, pipe, idx, transform)
                    else:
                        raise NotImplementedError(f"pipeline type of {type(pipeline)} is not supported")
                else:
                    logger.info("no pipeline in the data split")

            split = cfg.data.get(target)
            if split is not None:
                if isinstance(split, list):
                    for sub_item in split:
                        update_config(sub_item, pipeline_options)
                elif isinstance(split, dict):
                    update_config(split, pipeline_options)
                else:
                    logger.warning(f"type of split '{target}'' should be list or dict but {type(split)}")

        logger.info("configure_data()")
        logger.debug(f"[args] {cfg.data}")
        pipeline_options = cfg.data.pop("pipeline_options", None)
        if pipeline_options is not None and isinstance(pipeline_options, dict):
            configure_split("train")
            configure_split("val")
            if not training:
                configure_split("test")
            configure_split("unlabeled")

    def configure_ckpt(self, cfg, model_ckpt, pretrained=None):
        """Patch checkpoint path for pretrained weight.
        Replace cfg.load_from to model_ckpt
        Replace cfg.load_from to pretrained
        Replace cfg.resume_from to cfg.load_from
        """
        if model_ckpt:
            cfg.load_from = self.get_model_ckpt(model_ckpt)
        if pretrained and isinstance(pretrained, str):
            logger.info(f"Overriding cfg.load_from -> {pretrained}")
            cfg.load_from = pretrained  # Overriding by stage input
        if cfg.get("resume", False):
            cfg.resume_from = cfg.load_from

    @staticmethod
    def configure_hook(cfg, **kwargs):
        """Update cfg.custom_hooks based on cfg.custom_hook_options"""

        def update_hook(opt, custom_hooks, idx, hook):
            """Delete of update a custom hook"""
            if isinstance(opt, dict):
                if opt.get("_delete_", False):
                    # if option include _delete_=True, remove this hook from custom_hooks
                    logger.info(f"configure_hook: {hook['type']} is deleted")
                    del custom_hooks[idx]
                else:
                    logger.info(f"configure_hook: {hook['type']} is updated with {opt}")
                    hook.update(**opt)

        custom_hook_options = cfg.pop("custom_hook_options", {})
        # logger.info(f"configure_hook() {cfg.get('custom_hooks', [])} <- {custom_hook_options}")
        custom_hooks = cfg.get("custom_hooks", [])
        for idx, hook in enumerate(custom_hooks):
            for opt_key, opt in custom_hook_options.items():
                if hook["type"] == opt_key:
                    update_hook(opt, custom_hooks, idx, hook)

    @staticmethod
    def configure_samples_per_gpu(
        cfg: Config,
        subset: str,
        distributed: bool = False,
    ):

        dataloader_cfg = cfg.data.get(f"{subset}_dataloader", ConfigDict())
        samples_per_gpu = dataloader_cfg.get("samples_per_gpu", cfg.data.get("samples_per_gpu", 1))

        data_cfg = get_data_cfg(cfg, subset)
        dataset_len = len(data_cfg.otx_dataset)

        if distributed:
            dataset_len = dataset_len // dist.get_world_size()
        if dataset_len < samples_per_gpu:
            dataloader_cfg.samples_per_gpu = dataset_len
        cfg.data[f"{subset}_dataloader"] = dataloader_cfg

    @staticmethod
    def configure_compat_cfg(
        cfg: Config,
    ):
        """Modify config to keep the compatibility."""

        def _configure_dataloader(cfg):
            """Consume all the global dataloader config and convert them
            to specific dataloader config as it would be deprecated in the future.
            """
            global_dataloader_cfg = {}
            global_dataloader_cfg.update(
                {
                    k: cfg.data.pop(k)
                    for k in list(cfg.data.keys())
                    if k
                    not in [
                        "train",
                        "val",
                        "test",
                        "unlabeled",
                        "train_dataloader",
                        "val_dataloader",
                        "test_dataloader",
                        "unlabeled_dataloader",
                    ]
                }
            )

            for subset in ["train", "val", "test", "unlabeled"]:
                if subset not in cfg.data:
                    continue
                dataloader_cfg = cfg.data.get(f"{subset}_dataloader", None)
                if dataloader_cfg is None:
                    raise AttributeError(f"{subset}_dataloader is not found in config.")
                dataloader_cfg = {**global_dataloader_cfg, **dataloader_cfg}
                cfg.data[f"{subset}_dataloader"] = dataloader_cfg

        _configure_dataloader(cfg)

    @staticmethod
    def configure_fp16_optimizer(cfg: Config, distributed: bool = False):
        """Configure Fp16OptimizerHook and Fp16SAMOptimizerHook."""

        fp16_config = cfg.pop("fp16", None)
        if fp16_config is not None:
            optim_type = cfg.optimizer_config.get("type", "OptimizerHook")
            opts: Dict[str, Any] = dict(
                distributed=distributed,
                **fp16_config,
            )
            if optim_type == "SAMOptimizerHook":
                opts["type"] = "Fp16SAMOptimizerHook"
            elif optim_type == "OptimizerHook":
                opts["type"] = "Fp16OptimizerHook"
            else:
                # does not support optimizerhook type
                # let mm library handle it
                cfg.fp16 = fp16_config
                opts = dict()
            cfg.optimizer_config.update(opts)

    @staticmethod
    def configure_unlabeled_dataloader(cfg: Config, distributed: bool = False):
        if "unlabeled" in cfg.data:
            task_lib_module = importlib.import_module(f"{MODEL_TASK[cfg.model_task]}.datasets")
            dataset_builder = getattr(task_lib_module, "build_dataset")
            dataloader_builder = getattr(task_lib_module, "build_dataloader")

            dataset = build_dataset(cfg, "unlabeled", dataset_builder, consume=True)
            unlabeled_dataloader = build_dataloader(
                dataset,
                cfg,
                "unlabeled",
                dataloader_builder,
                distributed=distributed,
                consume=True,
            )

            custom_hooks = cfg.get("custom_hooks", [])
            updated = False
            for custom_hook in custom_hooks:
                if custom_hook["type"] == "ComposedDataLoadersHook":
                    custom_hook["data_loaders"] = [*custom_hook["data_loaders"], unlabeled_dataloader]
                    updated = True
            if not updated:
                custom_hooks.append(
                    ConfigDict(
                        type="ComposedDataLoadersHook",
                        data_loaders=unlabeled_dataloader,
                    )
                )
            cfg.custom_hooks = custom_hooks

    @staticmethod
    def get_model_meta(cfg):
        ckpt_path = cfg.get("load_from", None)
        meta = {}
        if ckpt_path:
            ckpt = CheckpointLoader.load_checkpoint(ckpt_path, map_location="cpu")
            meta = ckpt.get("meta", {})
        return meta

    @staticmethod
    def get_data_cfg(cfg, subset):
        assert subset in ["train", "val", "test"], f"Unknown subset:{subset}"
        if "dataset" in cfg.data[subset]:  # Concat|RepeatDataset
            dataset = cfg.data[subset].dataset
            while hasattr(dataset, "dataset"):
                dataset = dataset.dataset
            return dataset
        else:
            return cfg.data[subset]

    @staticmethod
    def get_data_classes(cfg):
        data_classes = []
        train_cfg = Stage.get_data_cfg(cfg, "train")
        if "data_classes" in train_cfg:
            data_classes = list(train_cfg.pop("data_classes", []))
        elif "classes" in train_cfg:
            data_classes = list(train_cfg.classes)
        return data_classes

    @staticmethod
    def get_model_classes(cfg):
        """Extract trained classes info from checkpoint file.
        MMCV-based models would save class info in ckpt['meta']['CLASSES']
        For other cases, try to get the info from cfg.model.classes (with pop())
        - Which means that model classes should be specified in model-cfg for
          non-MMCV models (e.g. OMZ models)
        """
        classes = []
        meta = Stage.get_model_meta(cfg)
        # for MPA classification legacy compatibility
        classes = meta.get("CLASSES", [])
        classes = meta.get("classes", classes)
        if classes is None:
            classes = []

        if len(classes) == 0:
            ckpt_path = cfg.get("load_from", None)
            if ckpt_path:
                classes = Stage.read_label_schema(ckpt_path)
        if len(classes) == 0:
            classes = cfg.model.pop("classes", cfg.pop("model_classes", []))
        return classes

    @staticmethod
    def get_model_ckpt(ckpt_path, new_path=None):
        ckpt = CheckpointLoader.load_checkpoint(ckpt_path, map_location="cpu")
        if "model" in ckpt:
            ckpt = ckpt["model"]
            if not new_path:
                new_path = ckpt_path[:-3] + "converted.pth"
            torch.save(ckpt, new_path)
            return new_path
        else:
            return ckpt_path

    @staticmethod
    def read_label_schema(ckpt_path, name_only=True, file_name="label_schema.json"):
        serialized_label_schema = []
        if any(ckpt_path.endswith(extension) for extension in (".xml", ".bin", ".pth")):
            label_schema_path = osp.join(osp.dirname(ckpt_path), file_name)
            if osp.exists(label_schema_path):
                with open(label_schema_path, encoding="UTF-8") as read_file:
                    serialized_label_schema = json.load(read_file)
        if serialized_label_schema:
            if name_only:
                all_classes = [labels["name"] for labels in serialized_label_schema["all_labels"].values()]
            else:
                all_classes = serialized_label_schema
        else:
            all_classes = []
        return all_classes

    @staticmethod
    def set_inference_progress_callback(model, cfg):
        # InferenceProgressCallback (Time Monitor enable into Infer task)
        time_monitor = None
        if cfg.get("custom_hooks", None):
            time_monitor = [hook.time_monitor for hook in cfg.custom_hooks if hook.type == "OTXProgressHook"]
            time_monitor = time_monitor[0] if time_monitor else None
        if time_monitor is not None:

            def pre_hook(module, input):
                time_monitor.on_test_batch_begin(None, None)

            def hook(module, input, output):
                time_monitor.on_test_batch_end(None, None)

            model.register_forward_pre_hook(pre_hook)
            model.register_forward_hook(hook)

    @classmethod
    def build_model(
        cls,
        cfg: Config,
        model_builder: Optional[Callable] = None,
        *,
        fp16: bool = False,
        **kwargs,
    ) -> torch.nn.Module:
        if model_builder is None:
            model_builder = cls.MODEL_BUILDER
        assert model_builder is not None
        model = model_builder(cfg, **kwargs)
        if bool(fp16):
            wrap_fp16_model(model)
        return model

    def _get_feature_module(self, model):
        return model



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from typing import Dict, List, Optional, Union

import openvino.runtime as ov
import torch

from .ov_model import OVModel
from .parser_mixin import ParserMixin


class MMOVModel(OVModel, ParserMixin):
    def __init__(
        self,
        model_path_or_model: Union[str, ov.Model],
        weight_path: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        outputs: Optional[Union[Dict[str, Union[str, List[str]]], List[str], str]] = None,
        *args,
        **kwargs
    ):
        parser = kwargs.pop("parser", None)
        parser_kwargs = kwargs.pop("parser_kwargs", {})
        inputs, outputs = super().parse(
            model_path_or_model=model_path_or_model,
            weight_path=weight_path,
            inputs=inputs,
            outputs=outputs,
            parser=parser,
            **parser_kwargs,
        )

        super().__init__(
            model_path_or_model=model_path_or_model,
            weight_path=weight_path,
            inputs=inputs,
            outputs=outputs,
            *args,
            **kwargs,
        )

    def forward(self, inputs, gt_label=None):
        if isinstance(inputs, torch.Tensor):
            inputs = (inputs,)
        assert len(inputs) == len(self.inputs)
        feed_dict = dict()
        for key, input in zip(self.inputs, inputs):
            feed_dict[key] = input

        if gt_label is not None:
            assert "gt_label" not in self.features
            self.features["gt_label"] = gt_label

        outputs = super().forward(**feed_dict)
        outputs = tuple(outputs.values())
        if len(outputs) == 1:
            outputs = outputs[0]
        return outputs



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .exporter import SemiSLDetectionExporter
from .inferrer import SemiSLDetectionInferrer
from .stage import SemiSLDetectionStage
from .trainer import SemiSLDetectionTrainer

__all__ = ["SemiSLDetectionStage", "SemiSLDetectionInferrer", "SemiSLDetectionTrainer", "SemiSLDetectionExporter"]



"""Configuration Manager ."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import shutil
from pathlib import Path
from typing import Any, Dict, List, Optional

from datumaro.components.dataset import Dataset
from datumaro.components.dataset_base import IDataset
from omegaconf import OmegaConf

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.helper import create
from otx.api.entities.model_template import ModelTemplate, parse_model_template
from otx.cli.registry import Registry as OTXRegistry
from otx.cli.utils.config import configure_dataset, override_parameters
from otx.cli.utils.importing import get_otx_root_path
from otx.cli.utils.parser import gen_param_help, gen_params_dict_from_args
from otx.core.data.manager.dataset_manager import DatasetManager

DEFAULT_MODEL_TEMPLATE_ID = {
    "CLASSIFICATION": "Custom_Image_Classification_EfficinetNet-B0",
    "DETECTION": "Custom_Object_Detection_Gen3_ATSS",
    "INSTANCE_SEGMENTATION": "Custom_Counting_Instance_Segmentation_MaskRCNN_ResNet50",
    "ROTATED_DETECTION": "Custom_Rotated_Detection_via_Instance_Segmentation_MaskRCNN_ResNet50",
    "SEGMENTATION": "Custom_Semantic_Segmentation_Lite-HRNet-18-mod2_OCR",
    "ACTION_CLASSIFICATION": "Custom_Action_Classificaiton_X3D",
    "ACTION_DETECTION": "Custom_Action_Detection_X3D_FAST_RCNN",
    "ANOMALY_CLASSIFICATION": "ote_anomaly_classification_padim",
    "ANOMALY_DETECTION": "ote_anomaly_detection_padim",
    "ANOMALY_SEGMENTATION": "ote_anomaly_segmentation_padim",
}

AUTOSPLIT_SUPPORTED_FORMAT = [
    "imagenet",
    "coco",
    "cityscapes",
    "voc",
]

TASK_TYPE_TO_SUPPORTED_FORMAT = {
    "CLASSIFICATION": ["imagenet", "datumaro"],
    "DETECTION": ["coco", "voc", "yolo"],
    "SEGMENTATION": ["cityscapes", "common_semantic_segmentation", "voc", "ade20k2017", "ade20k2020"],
    "ACTION_CLASSIFICATION": ["multi-cvat"],
    "ACTION_DETECTION": ["multi-cvat"],
    "ANOMALY_CLASSIFICATION": ["mvtec"],
    "ANOMALY_DETECTION": ["mvtec"],
    "ANOMALY_SEGMENTATION": ["mvtec"],
    "INSTANCE_SEGMENTATION": ["coco", "voc"],
    "ROTATED_DETECTION": ["coco", "voc"],
}

TASK_TYPE_TO_SUB_DIR_NAME = {
    "INCREMENTAL": "",
    "SEMISUPERVISED": "semisl",
    "SELFSUPERVISED": "selfsl",
}


def set_workspace(task: str, root: str = None, name: str = "otx-workspace"):
    """Set workspace path according to arguments."""
    path = f"{root}/{name}-{task}" if root else f"./{name}-{task}"
    return path


class ConfigManager:  # pylint: disable=too-many-instance-attributes
    """Auto configuration manager that could set the proper configuration.

    Currently, it only supports the small amount of functions.
    * Data format detection
    * Task type detection
    * Write the data to the workspace
    * Write the data configuration to the workspace

    However, it will supports lots of things in the near future.
    * Automatic train type detection (Supervised, Self, Semi)
    * Automatic resource allocation (num_workers, HPO)

    """

    def __init__(self, args, workspace_root: Optional[str] = None, mode: str = "train"):
        # Currently, Datumaro.auto_split() can support below 3 tasks
        # Classification, Detection, Segmentation
        self.otx_root = get_otx_root_path()
        self.workspace_root = Path(workspace_root) if workspace_root else Path(".")
        self.mode = mode
        self.rebuild: bool = False

        self.args = args
        self.template = args.template
        self.task_type: str = ""
        self.train_type: str = ""
        self.model: str = ""

        self.dataset_manager = DatasetManager()
        self.data_format: str = ""
        self.data_config: Dict[str, dict] = {}

    @property
    def data_config_file_path(self) -> Path:
        """The path of the data configuration yaml to use for the task.

        Raises:
            FileNotFoundError: If data is received as args from otx train and the file does not exist, Error.

        Returns:
            Path: Path of target data configuration file.
        """
        if "data" in self.args and self.args.data:
            if Path(self.args.data).exists():
                return Path(self.args.data)
            raise FileNotFoundError(f"Not found: {self.args.data}")
        return self.workspace_root / "data.yaml"

    def check_workspace(self) -> bool:
        """Check that the class's workspace_root is an actual workspace folder.

        Returns:
            bool: true for workspace else false
        """
        has_template_yaml = (self.workspace_root / "template.yaml").exists()
        has_data_yaml = self.data_config_file_path.exists()
        return has_template_yaml and has_data_yaml

    def configure_template(self, model: str = None) -> None:
        """Update the template appropriate for the situation."""
        if self.check_workspace():
            # Workspace -> template O
            self.template = parse_model_template(str(self.workspace_root / "template.yaml"))
            if self.mode == "build" and self._check_rebuild():
                self.rebuild = True
                model = model if model else self.template.name
                self.template = self._get_template(str(self.task_type), model=model)
        elif self.template and Path(self.template).exists():
            # No workspace -> template O
            self.template = parse_model_template(self.template)
        else:
            task_type = self.task_type
            if not task_type and not model:
                task_type = self.auto_task_detection(self.args.train_data_roots)
            self.template = self._get_template(task_type, model=model)
        self.task_type = self.template.task_type
        self.model = self.template.name
        self.train_type = self._get_train_type()

    def _check_rebuild(self):
        """Checking for Rebuild status."""
        if self.args.task and str(self.template.task_type) != self.args.task.upper():
            raise NotImplementedError("Task Update is not yet supported.")
        result = False
        if self.args.model and self.template.name != self.args.model.upper():
            print(f"[*] Rebuild model: {self.template.name} -> {self.args.model.upper()}")
            result = True
        template_train_type = self._get_train_type(ignore_args=True)
        if self.args.train_type and template_train_type != self.args.train_type.upper():
            self.train_type = self.args.train_type.upper()
            print(f"[*] Rebuild train-type: {template_train_type} -> {self.train_type}")
            result = True
        return result

    def configure_data_config(self, update_data_yaml: bool = True) -> None:
        """Configure data_config according to the situation and create data.yaml."""
        data_yaml_path = self.data_config_file_path
        data_yaml = configure_dataset(self.args, data_yaml_path=data_yaml_path)
        if self.mode in ("train", "build"):
            use_auto_split = data_yaml["data"]["train"]["data-roots"] and not data_yaml["data"]["val"]["data-roots"]
            # FIXME: Hardcoded for Self-Supervised Learning
            if use_auto_split and str(self.train_type).upper() != "SELFSUPERVISED":
                splitted_dataset = self.auto_split_data(data_yaml["data"]["train"]["data-roots"], str(self.task_type))
                default_data_folder_name = "splitted_dataset"
                data_yaml = self._get_arg_data_yaml()
                self._save_data(splitted_dataset, default_data_folder_name, data_yaml)
        if update_data_yaml:
            self._export_data_cfg(data_yaml, str(data_yaml_path))
            print(f"[*] Update data configuration file to: {str(data_yaml_path)}")
        self.update_data_config(data_yaml)

    def _get_train_type(self, ignore_args: bool = False) -> str:
        """Check and return the train_type received as input args."""
        if not ignore_args:
            args_hyper_parameters = gen_params_dict_from_args(self.args)
            arg_algo_backend = args_hyper_parameters.get("algo_backend", False)
            if arg_algo_backend:
                train_type = arg_algo_backend.get("train_type", {"value": "INCREMENTAL"})  # type: ignore
                return train_type.get("value", "INCREMENTAL")
            if self.mode in ("build") and self.args.train_type:
                self.train_type = self.args.train_type.upper()
            if self.train_type in TASK_TYPE_TO_SUB_DIR_NAME:
                return self.train_type

        algo_backend = self.template.hyper_parameters.parameter_overrides.get("algo_backend", False)
        if algo_backend:
            train_type = algo_backend.get("train_type", {"default_value": "INCREMENTAL"})
            return train_type.get("default_value", "INCREMENTAL")
        return "INCREMENTAL"

    def auto_task_detection(self, data_roots: str) -> str:
        """Detect task type automatically."""
        if not data_roots:
            raise ValueError("Workspace must already exist or one of {task or model or train-data-roots} must exist.")
        self.data_format = self.dataset_manager.get_data_format(data_roots)
        return self._get_task_type_from_data_format(self.data_format)

    def _get_task_type_from_data_format(self, data_format: str) -> str:
        """Detect task type.

        For some datasets (i.e. COCO, VOC, MVTec), can't be fully automated.
        Because those datasets have several format at the same time.
        (i.e. for the COCO case, object detection and instance segmentation annotations coexist)
        In this case, the task_type will be selected to default value.

        For action tasks, currently action_classification is default.

        If Datumaro supports the Kinetics, AVA datasets, MVTec, _is_cvat_format(), _is_mvtec_format()
        functions will be deleted.
        """

        for task_key, data_value in TASK_TYPE_TO_SUPPORTED_FORMAT.items():
            if data_format in data_value:
                self.task_type = task_key
                print(f"[*] Detected task type: {self.task_type}")
                return task_key
        raise ValueError(f"Can't find proper task. we are not support {data_format} format, yet.")

    def auto_split_data(self, data_roots: str, task: str):
        """Automatically Split train data --> train/val dataset."""
        self.data_format = self.dataset_manager.get_data_format(data_roots)
        dataset = self.dataset_manager.import_dataset(data_root=data_roots, data_format=self.data_format)
        train_dataset = self.dataset_manager.get_train_dataset(dataset)
        val_dataset = self.dataset_manager.get_val_dataset(dataset)
        splitted_dataset = None
        if self.data_format in AUTOSPLIT_SUPPORTED_FORMAT:
            if val_dataset is None:
                splitted_dataset = self.dataset_manager.auto_split(
                    task=task,
                    dataset=train_dataset,
                    split_ratio=[("train", 0.8), ("val", 0.2)],
                )
            else:
                print(f"[*] Found validation data in your dataset in {data_roots}. It'll be used as validation data.")
                splitted_dataset = {"train": train_dataset, "val": val_dataset}
        else:
            print(f"[*] Current auto-split can't support the {self.data_format} format.")
        return splitted_dataset

    def _get_arg_data_yaml(self):
        # TODO: This should modify data yaml format to data_config format.
        """Save the splitted dataset and data.yaml to the workspace."""
        data_yaml = self._create_empty_data_cfg()
        if self.mode == "train":
            if self.args.train_data_roots:
                data_yaml["data"]["train"]["data-roots"] = self.args.train_data_roots
            if self.args.val_data_roots:
                data_yaml["data"]["val"]["data-roots"] = self.args.val_data_roots
            if self.args.unlabeled_data_roots:
                data_yaml["data"]["unlabeled"]["data-roots"] = self.args.unlabeled_data_roots
        elif self.mode == "test":
            if self.args.test_data_roots:
                data_yaml["data"]["test"]["data-roots"] = self.args.test_data_roots
        return data_yaml

    def _save_data(
        self,
        splitted_dataset: Dict[str, IDataset],
        default_data_folder_name: str,
        data_config: Dict[str, Dict[str, Dict[str, Any]]],
    ) -> None:
        """Save the data for the classification task.

        Args:
            splitted_dataset (dict): A dictionary containing split datasets
            default_data_folder_name (str): the name of splitted dataset folder
            data_config (dict): dictionary that has information about data path
        """
        for phase, dataset in splitted_dataset.items():
            dst_dir_path = self.workspace_root / default_data_folder_name / phase
            data_config["data"][phase]["data-roots"] = str(dst_dir_path.absolute())
            # Convert Datumaro class: DatasetFilter(IDataset) --> Dataset
            if isinstance(dataset, Dataset):
                datum_dataset = dataset
            else:
                datum_dataset = Dataset.from_extractors(dataset)
            # Write the data
            # TODO: consider the way that reduces disk stroage
            # Currently, saving all images to the workspace.
            # It might needs quite large disk storage.
            self.dataset_manager.export_dataset(
                dataset=datum_dataset, output_dir=str(dst_dir_path), data_format=self.data_format, save_media=True
            )

        if data_config["data"]["unlabeled"]["data-roots"] is not None:
            data_config["data"]["unlabeled"]["data-roots"] = str(
                Path(data_config["data"]["unlabeled"]["data-roots"]).absolute()
            )

    def _create_empty_data_cfg(self) -> Dict[str, Dict[str, Dict[str, Any]]]:
        """Create default dictionary to represent the dataset."""
        data_config: Dict[str, Dict[str, Any]] = {"data": {}}
        for subset in ["train", "val", "test"]:
            data_subset = {"ann-files": None, "data-roots": None}
            data_config["data"][subset] = data_subset
        data_config["data"]["unlabeled"] = {"file-list": None, "data-roots": None}
        return data_config

    def _export_data_cfg(self, data_cfg: Dict[str, Dict[str, Dict[str, Any]]], output_path: str) -> None:
        """Export the data configuration file to output_path."""
        Path(output_path).write_text(OmegaConf.to_yaml(data_cfg), encoding="utf-8")

    def get_hyparams_config(self, override_param: Optional[List] = None) -> ConfigurableParameters:
        """Separates the input params received from args and updates them.."""
        hyper_parameters = self.template.hyper_parameters.data
        type_hint = gen_param_help(hyper_parameters)
        updated_hyper_parameters = gen_params_dict_from_args(
            self.args, override_param=override_param, type_hint=type_hint
        )
        override_parameters(updated_hyper_parameters, hyper_parameters)
        return create(hyper_parameters)

    def get_dataset_config(self, subsets: List[str]) -> dict:
        """Returns dataset_config in a format suitable for each subset.

        Args:
            subsets (list, str): Defaults to ["train", "val", "unlabeled"].

        Returns:
            dict: dataset_config
        """
        dataset_config = {"task_type": self.task_type, "train_type": self.train_type}
        for subset in subsets:
            if f"{subset}_subset" in self.data_config and self.data_config[f"{subset}_subset"]["data_root"]:
                dataset_config.update({f"{subset}_data_roots": self.data_config[f"{subset}_subset"]["data_root"]})
        return dataset_config

    def update_data_config(self, data_yaml: dict) -> None:
        # TODO: This also requires uniformity in the format.
        """Convert the data yaml format to the data_config format consumed by the task.

        Args:
            data_yaml (dict): data.yaml format
        """
        if data_yaml["data"]["train"]["data-roots"]:
            self.data_config["train_subset"] = {"data_root": data_yaml["data"]["train"]["data-roots"]}
        if data_yaml["data"]["val"]["data-roots"]:
            self.data_config["val_subset"] = {"data_root": data_yaml["data"]["val"]["data-roots"]}
        if data_yaml["data"]["test"]["data-roots"]:
            self.data_config["test_subset"] = {"data_root": data_yaml["data"]["test"]["data-roots"]}
        if "unlabeled" in data_yaml["data"] and data_yaml["data"]["unlabeled"]["data-roots"]:
            self.data_config["unlabeled_subset"] = {
                "data_root": data_yaml["data"]["unlabeled"]["data-roots"],
                "file_list": data_yaml["data"]["unlabeled"]["file-list"],
            }
        # FIXME: Hardcoded for Self-Supervised Learning
        if self.mode == "train" and str(self.train_type).upper() == "SELFSUPERVISED":
            self.data_config["val_subset"] = {"data_root": None}

    def _get_template(self, task_type: str, model: Optional[str] = None) -> ModelTemplate:
        """Returns the appropriate template for each situation.

        Args:
            task_type (str): The task_type registered in the registry. Used for filtering.
            model (str, optional): The task_type registered in the registry. Used for filtering. Defaults to None.

        Returns:
            ModelTemplate: Selected model template.
        """
        otx_registry = OTXRegistry(self.otx_root).filter(task_type=task_type if task_type else None)
        if model:
            template_lst = [temp for temp in otx_registry.templates if temp.name.lower() == model.lower()]
            if not template_lst:
                raise ValueError(
                    f"[*] {model} is not a type supported by OTX {task_type}."
                    f"\n[*] Please refer to 'otx find --template --task_type {task_type}'"
                )
            template = template_lst[0]
        else:
            template = otx_registry.get(DEFAULT_MODEL_TEMPLATE_ID[task_type.upper()])
        return template

    def build_workspace(self, new_workspace_path: Optional[str] = None) -> None:
        """Create OTX workspace with Template configs from task type.

        This function provides a user-friendly OTX workspace and provides more intuitive
        and create customizable templates to help users use all the features of OTX.

        Args:
            new_workspace_path (Optional[str]): Workspace dir name for build
        """

        # Create OTX-workspace
        # Check whether the workspace is existed or not
        if self.check_workspace() and not self.rebuild:
            return
        if self.rebuild:
            print(f"[*] \t- Rebuild: model-{self.model} / train type-{self.train_type}")
        if new_workspace_path:
            self.workspace_root = Path(new_workspace_path)
        elif not self.check_workspace():
            self.workspace_root = Path(set_workspace(task=self.task_type))
        self.workspace_root.mkdir(exist_ok=True, parents=True)
        print(f"[*] Workspace Path: {self.workspace_root}")
        print(f"[*] Load Model Template ID: {self.template.model_template_id}")
        print(f"[*] Load Model Name: {self.template.name}")

        template_dir = Path(self.template.model_template_path).parent

        # Copy task base configuration file
        task_configuration_path = template_dir / self.template.hyper_parameters.base_path
        shutil.copyfile(task_configuration_path, str(self.workspace_root / "configuration.yaml"))
        # Load Model Template
        template_config = OmegaConf.load(self.template.model_template_path)
        template_config.hyper_parameters.base_path = "./configuration.yaml"

        # Configuration of Train Type value
        train_type_rel_path = TASK_TYPE_TO_SUB_DIR_NAME[self.train_type]

        # FIXME: Hardcoded solution for supcon
        enable_supcon = gen_params_dict_from_args(self.args).get("learning_parameters", {})
        enable_supcon = enable_supcon.get("enable_supcon", {"value": False})
        if enable_supcon.get("value", False):
            train_type_rel_path = "supcon"

        model_dir = template_dir.absolute() / train_type_rel_path
        if not model_dir.exists():
            raise ValueError(f"[*] {self.train_type} is not a type supported by OTX {self.task_type}")
        train_type_dir = self.workspace_root / train_type_rel_path
        train_type_dir.mkdir(exist_ok=True)

        # Update Hparams
        if (model_dir / "hparam.yaml").exists():
            template_config = OmegaConf.merge(template_config, OmegaConf.load(str(model_dir / "hparam.yaml")))

        # Copy config files
        config_files = [
            (model_dir, "model.py", train_type_dir),
            (model_dir, "data_pipeline.py", train_type_dir),
            (template_dir, "tile_pipeline.py", self.workspace_root),
            (template_dir, "deployment.py", self.workspace_root),
            (template_dir, "hpo_config.yaml", self.workspace_root),
            (template_dir, "model_hierarchical.py", self.workspace_root),
            (template_dir, "model_multilabel.py", self.workspace_root),
        ]
        for target_dir, file_name, dest_dir in config_files:
            self._copy_config_files(target_dir, file_name, dest_dir)
        (self.workspace_root / "template.yaml").write_text(OmegaConf.to_yaml(template_config))

        # Copy compression_config.json
        if (model_dir / "compression_config.json").exists():
            shutil.copyfile(
                str(model_dir / "compression_config.json"),
                str(train_type_dir / "compression_config.json"),
            )
            print(f"[*] \t- Updated: {str(train_type_dir / 'compression_config.json')}")
        # Copy compression_config.json
        if (model_dir / "pot_optimization_config.json").exists():
            shutil.copyfile(
                str(model_dir / "pot_optimization_config.json"),
                str(train_type_dir / "pot_optimization_config.json"),
            )
            print(f"[*] \t- Updated: {str(train_type_dir / 'pot_optimization_config.json')}")

        if not (self.workspace_root / "data.yaml").exists():
            data_yaml = self._get_arg_data_yaml()
            self._export_data_cfg(data_yaml, str((self.workspace_root / "data.yaml")))

        self.template = parse_model_template(str(self.workspace_root / "template.yaml"))

    def _copy_config_files(self, target_dir: Path, file_name: str, dest_dir: Path) -> None:
        """Copy Configuration files for workspace."""
        if (target_dir / file_name).exists():
            if file_name.endswith(".py"):
                try:
                    from otx.mpa.utils.config_utils import MPAConfig

                    config = MPAConfig.fromfile(str(target_dir / file_name))
                    config.dump(str(dest_dir / file_name))
                except Exception as exc:
                    raise ImportError(f"{self.task_type} requires mmcv-full to be installed.") from exc
            elif file_name.endswith((".yml", ".yaml")):
                config = OmegaConf.load(str(target_dir / file_name))
                (dest_dir / file_name).write_text(OmegaConf.to_yaml(config))
            print(f"[*] \t- Updated: {str(dest_dir / file_name)}")



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import otx.mpa.modules.datasets.pipelines.compose
import otx.mpa.modules.datasets.pipelines.transforms.seg_custom_pipelines
import otx.mpa.modules.hooks
import otx.mpa.modules.models.backbones.litehrnet
import otx.mpa.modules.models.heads.custom_fcn_head
import otx.mpa.modules.models.losses.cross_entropy_loss_with_ignore
import otx.mpa.modules.models.scalar_schedulers.constant
import otx.mpa.modules.models.scalar_schedulers.poly
import otx.mpa.modules.models.scalar_schedulers.step
import otx.mpa.modules.models.segmentors
from otx.mpa.seg.incremental import IncrSegInferrer, IncrSegTrainer
from otx.mpa.seg.semisl import SemiSLSegExporter, SemiSLSegInferrer, SemiSLSegTrainer

# flake8: noqa
from . import exporter, inferrer, stage, trainer



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmcv.runner import HOOKS

from otx.mpa.utils.logger import get_logger

from .model_ema_hook import DualModelEMAHook

logger = get_logger()


@HOOKS.register_module()
class UnbiasedTeacherHook(DualModelEMAHook):
    def __init__(self, min_pseudo_label_ratio=0.1, **kwargs):
        super().__init__(**kwargs)
        self.min_pseudo_label_ratio = min_pseudo_label_ratio
        self.unlabeled_loss_enabled = False

    def before_train_epoch(self, runner):
        super().before_train_epoch(runner)

        if runner.epoch + 1 < self.start_epoch:
            return
        if self.unlabeled_loss_enabled:
            return

        average_pseudo_label_ratio = self._get_average_pseudo_label_ratio(runner)
        logger.info(f"avr_ps_ratio: {average_pseudo_label_ratio}")
        if average_pseudo_label_ratio > self.min_pseudo_label_ratio:
            self._get_model(runner).enable_unlabeled_loss()
            self.unlabeled_loss_enabled = True
            logger.info("---------- Enabled unlabeled loss")

    def after_train_iter(self, runner):
        """Update ema parameter every self.interval iterations."""
        if runner.iter % self.interval != 0:
            # Skip update
            return

        if runner.epoch + 1 < self.start_epoch or self.unlabeled_loss_enabled is False:
            # Just copy parameters before enabled
            self._copy_model()
            return

        # EMA
        self._ema_model()

    def _get_average_pseudo_label_ratio(self, runner):
        output_backup = runner.log_buffer.output.copy()
        was_ready = runner.log_buffer.ready
        runner.log_buffer.average(100)
        average_pseudo_label_ratio = runner.log_buffer.output.get("ps_ratio", 0.0)
        runner.log_buffer.output = output_backup
        runner.ready = was_ready
        return average_pseudo_label_ratio



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest
import torch
from torch import nn

from otx.algorithms.common.adapters.nncf.compression import (
    AccuracyAwareLrUpdater,
    NNCFMetaState,
    get_uncompressed_model,
    is_checkpoint_nncf,
    is_state_nncf,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def helper(fn, returns_with_params):
    for item in returns_with_params:
        ret = item[0]
        args = item[1]
        kwargs = item[2]

        assert ret == fn(*args, **kwargs)


class TestNNCFMetaState:
    @e2e_pytest_unit
    def test_repr(self):
        state = NNCFMetaState(None, None, None)
        assert repr(state) == "NNCFMetaState()"
        assert state.state_to_build is None
        assert state.data_to_build is None
        assert state.compression_ctrl is None

        state = NNCFMetaState({"dummy": torch.tensor(1)})
        assert repr(state) == "NNCFMetaState(state_to_build='<data>')"
        assert state.state_to_build == {"dummy": torch.tensor(1)}
        assert state.data_to_build is None
        assert state.compression_ctrl is None

        state = NNCFMetaState(None, np.array(1))
        assert repr(state) == "NNCFMetaState(data_to_build='<data>')"
        assert state.state_to_build is None
        assert state.data_to_build == np.array(1)
        assert state.compression_ctrl is None

        state = NNCFMetaState(None, None, {"dummy": "dummy"})
        assert repr(state) == "NNCFMetaState(compression_ctrl='<data>')"
        assert state.state_to_build is None
        assert state.data_to_build is None
        assert state.compression_ctrl == {"dummy": "dummy"}


@e2e_pytest_unit
def test_is_state_nncf():
    returns_with_params = [
        (True, [{"meta": {"nncf_enable_compression": True}}], {}),
        (False, [{"meta": {"nncf_enable_compression": False}}], {}),
        (False, [{"meta": {}}], {}),
    ]

    helper(is_state_nncf, returns_with_params)


@e2e_pytest_unit
def test_is_checkpoint_nncf():
    returns_with_params = [
        (False, ["dummy_file_path"], {}),
    ]

    helper(is_checkpoint_nncf, returns_with_params)


@pytest.fixture()
def mock_model():
    class MockModule(nn.Module):
        def __init__(self):
            super().__init__()

    return MockModule()


@e2e_pytest_unit
def test_get_uncompressed_model(mock_model):
    returns_with_params = [
        (mock_model, [mock_model], {}),
    ]

    helper(get_uncompressed_model, returns_with_params)


class TestAccuracyAwareLrUpdater:
    @e2e_pytest_unit
    def test_step(self, mock_model):
        updater = AccuracyAwareLrUpdater(mock_model)
        updater.step()

    @e2e_pytest_unit
    def test_base_lrs(self, mock_model):
        updater = AccuracyAwareLrUpdater(mock_model)
        for value in np.arange(0, 0.1, 0.005):
            updater.base_lrs = value
            assert value == updater._lr_hook.base_lr



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .constant import ConstantScalarScheduler
from .poly import PolyScalarScheduler
from .step import StepScalarScheduler

__all__ = [
    "ConstantScalarScheduler",
    "PolyScalarScheduler",
    "StepScalarScheduler",
]



img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
__resize_target_size = 224


train_pipeline = [
    dict(
        type="TwoCropTransform",
        pipeline=[
            dict(type="Resize", size=__resize_target_size),
            dict(type="RandomFlip", flip_prob=0.5, direction="horizontal"),
            dict(type="AugMixAugment", config_str="augmix-m5-w3"),
            dict(type="RandomRotate", p=0.35, angle=(-10, 10)),
            dict(type="ToNumpy"),
            dict(type="Normalize", **img_norm_cfg),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="ToTensor", keys=["gt_label"]),
            dict(type="Collect", keys=["img", "gt_label"]),
        ],
    )
]

test_pipeline = [
    dict(type="Resize", size=__resize_target_size),
    dict(type="Normalize", **img_norm_cfg),
    dict(type="ImageToTensor", keys=["img"]),
    dict(type="Collect", keys=["img"]),
]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

DOMAIN_CUSTOM_OPS_NAME = "org.openvinotoolkit"


def add_domain(name_operator: str) -> str:
    return DOMAIN_CUSTOM_OPS_NAME + "::" + name_operator



"""BaseTask for Classification/Detection/Segmentation."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.


import abc
import io
import os
import shutil
import tempfile
from copy import deepcopy
from typing import Dict, List, Optional, Union

import numpy as np
import torch
from mmcv.utils.config import Config, ConfigDict

from otx.algorithms.common.adapters.mmcv.hooks import OTXLoggerHook
from otx.algorithms.common.adapters.mmcv.utils import (
    align_data_config_with_recipe,
    get_configs_by_pairs,
)
from otx.algorithms.common.configs import TrainType
from otx.algorithms.common.utils import UncopiableDefaultDict
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label import LabelEntity
from otx.api.entities.model import ModelEntity, ModelPrecision, OptimizationMethod
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.serialization.label_mapper import LabelSchemaMapper
from otx.api.usecases.reporting.time_monitor_callback import TimeMonitorCallback
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.export_interface import IExportTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.unload_interface import IUnload
from otx.api.utils.argument_checks import check_input_parameters_type
from otx.mpa.builder import build
from otx.mpa.modules.hooks.cancel_interface_hook import CancelInterfaceHook
from otx.mpa.stage import Stage
from otx.mpa.utils.config_utils import (
    MPAConfig,
    add_custom_hook_if_not_exists,
    remove_custom_hook,
    update_or_add_custom_hook,
)
from otx.mpa.utils.logger import get_logger

logger = get_logger()
TRAIN_TYPE_DIR_PATH = {
    TrainType.INCREMENTAL.name: ".",
    TrainType.SELFSUPERVISED.name: "selfsl",
    TrainType.SEMISUPERVISED.name: "semisl",
}


# pylint: disable=too-many-instance-attributes, protected-access
class BaseTask(IInferenceTask, IExportTask, IEvaluationTask, IUnload):
    """BaseTask for OTX Algorithms."""

    _task_environment: TaskEnvironment

    @check_input_parameters_type()
    def __init__(self, task_config, task_environment: TaskEnvironment, output_path: Optional[str] = None):
        self._task_config = task_config
        self._task_environment = task_environment
        self._hyperparams = task_environment.get_hyper_parameters(self._task_config)  # type: ConfigDict
        self._model_name = task_environment.model_template.name
        self._task_type = task_environment.model_template.task_type
        self._labels = task_environment.get_labels(include_empty=False)
        self.confidence_threshold = self._get_confidence_threshold(self._hyperparams)
        # Set default model attributes.
        self._model_label_schema = []  # type: List[LabelEntity]
        self._optimization_methods = []  # type: List[OptimizationMethod]
        self._model_ckpt = None
        self._resume = False
        self._anchors = {}  # type: Dict[str, int]
        self._work_dir_is_temp = False
        if output_path is None:
            output_path = tempfile.mkdtemp(prefix="OTX-task-")
            self._work_dir_is_temp = True
        self._output_path = output_path
        logger.info(f"created output path at {self._output_path}")
        if task_environment.model is not None:
            logger.info("loading the model from the task env.")
            state_dict = self._load_model_ckpt(self._task_environment.model)
            if state_dict:
                self._model_ckpt = os.path.join(self._output_path, "env_model_ckpt.pth")
                if os.path.exists(self._model_ckpt):
                    os.remove(self._model_ckpt)
                torch.save(state_dict, self._model_ckpt)
                self._model_label_schema = self._load_model_label_schema(self._task_environment.model)
                self._resume = self._load_resume_info(self._task_environment.model)

        # property below will be initialized by initialize()
        self._recipe_cfg = None
        self._stage_module = None
        self._model_cfg = None
        self._precision = [ModelPrecision.FP32]
        self._data_cfg = None
        self._mode = None
        self._time_monitor = None  # type: Optional[TimeMonitorCallback]
        self._learning_curves = UncopiableDefaultDict(OTXLoggerHook.Curve)
        self._is_training = False
        self._should_stop = False
        self.cancel_interface = None
        self.reserved_cancel = False
        self.on_hook_initialized = self.OnHookInitialized(self)

        # Initialize Train type related var
        self._train_type = self._hyperparams.algo_backend.train_type
        self._model_dir = os.path.join(
            os.path.abspath(os.path.dirname(self.template_file_path)), TRAIN_TYPE_DIR_PATH[self._train_type.name]
        )

        # to override configuration at runtime
        self.override_configs = {}  # type: Dict[str, str]

    def _run_task(self, stage_module, mode=None, dataset=None, **kwargs):
        self._initialize(kwargs)
        stage_module = self._update_stage_module(stage_module)

        if mode is not None:
            self._mode = mode

        # deepcopy all configs to make sure
        # changes under MPA and below does not take an effect to OTX for clear distinction
        model_cfg = deepcopy(self._model_cfg)
        recipe_cfg = deepcopy(self._recipe_cfg)
        data_cfg = deepcopy(self._data_cfg)
        assert model_cfg is not None, "'model_cfg' is not initialized."
        assert recipe_cfg is not None, "'recipe_cfg' is not initialized."

        # update model config -> model label schema
        data_classes = [label.name for label in self._labels]
        model_classes = [label.name for label in self._model_label_schema]
        model_cfg["model_classes"] = model_classes
        if dataset is not None:
            train_data_cfg = Stage.get_data_cfg(data_cfg, "train")
            train_data_cfg["data_classes"] = data_classes
            new_classes = np.setdiff1d(data_classes, model_classes).tolist()
            train_data_cfg["new_classes"] = new_classes

        logger.info(  # pylint: disable=logging-not-lazy
            "running task... kwargs = "
            + str({k: v if k != "model_builder" else object.__repr__(v) for k, v in kwargs.items()})
        )

        common_cfg = ConfigDict(dict(output_path=self._output_path, resume=self._resume))

        # build workflow using recipe configuration
        workflow = build(
            recipe_cfg,
            self._mode,
            stage_type=stage_module,
            common_cfg=common_cfg,
        )

        # run workflow with task specific model config and data config
        output = workflow.run(
            model_cfg=model_cfg,
            data_cfg=data_cfg,
            ir_model_path=None,
            ir_weight_path=None,
            model_ckpt=self._model_ckpt,
            mode=self._mode,
            **kwargs,
        )
        logger.info("run task done.")
        return output

    def _delete_scratch_space(self):
        """Remove model checkpoints and mpa logs."""
        if os.path.exists(self._output_path):
            shutil.rmtree(self._output_path, ignore_errors=False)

    def _pre_task_run(self):
        pass

    @property
    def project_path(self):
        """Return output path with logs."""
        return self._output_path

    @property
    def model_name(self):
        """Name of Model Template."""
        return self._task_environment.model_template.name

    @property
    def labels(self):
        """Label List of Task Environment."""
        return self._task_environment.get_labels(False)

    @property
    def template_file_path(self):
        """Model Template file path."""
        return self._task_environment.model_template.model_template_path

    @property
    def data_pipeline_path(self):
        """Base Data Pipeline file path."""
        # TODO: Temporarily use data_pipeline.py next to model.py.may change later.
        if self._hyperparams.tiling_parameters.enable_tiling:
            return os.path.join(self._model_dir, "tile_pipeline.py")
        return os.path.join(self._model_dir, "data_pipeline.py")

    @property
    def hyperparams(self):
        """Hyper Parameters configuration."""
        return self._hyperparams

    # pylint: disable-next=too-many-branches,too-many-statements
    def _initialize(self, options=None):  # noqa: C901
        """Prepare configurations to run a task through MPA's stage."""
        if options is None:
            options = {}

        export = options.get("export", False)

        logger.info("initializing....")
        self._init_recipe()

        if not export:
            # FIXME: Temporary remedy for CVS-88098
            recipe_hparams = self._init_recipe_hparam()
            if len(recipe_hparams) > 0:
                self._recipe_cfg.merge_from_dict(recipe_hparams)

        if "custom_hooks" in self.override_configs:
            override_custom_hooks = self.override_configs.pop("custom_hooks")
            for override_custom_hook in override_custom_hooks:
                update_or_add_custom_hook(self._recipe_cfg, ConfigDict(override_custom_hook))
        if len(self.override_configs) > 0:
            logger.info(f"before override configs merging = {self._recipe_cfg}")
            self._recipe_cfg.merge_from_dict(self.override_configs)
            logger.info(f"after override configs merging = {self._recipe_cfg}")

        # prepare model config
        self._model_cfg = self._init_model_cfg()

        # Remove FP16 config if running on CPU device and revert to FP32
        # https://github.com/pytorch/pytorch/issues/23377
        if not torch.cuda.is_available() and "fp16" in self._model_cfg:
            logger.info("Revert FP16 to FP32 on CPU device")
            if isinstance(self._model_cfg, Config):
                del self._model_cfg._cfg_dict["fp16"]
            elif isinstance(self._model_cfg, ConfigDict):
                del self._model_cfg["fp16"]

        # default adaptive hook for evaluating before and after training
        add_custom_hook_if_not_exists(
            self._recipe_cfg,
            ConfigDict(
                type="AdaptiveTrainSchedulingHook",
                enable_adaptive_interval_hook=False,
                enable_eval_before_run=True,
            ),
        )
        # Add/remove adaptive interval hook
        if self._recipe_cfg.get("use_adaptive_interval", False):
            update_or_add_custom_hook(
                self._recipe_cfg,
                ConfigDict(
                    {
                        "type": "AdaptiveTrainSchedulingHook",
                        "max_interval": 5,
                        "enable_adaptive_interval_hook": True,
                        "enable_eval_before_run": True,
                        **self._recipe_cfg.pop("adaptive_validation_interval", {}),
                    }
                ),
            )
        else:
            self._recipe_cfg.pop("adaptive_validation_interval", None)

        self.set_early_stopping_hook()

        # add Cancel tranining hook
        update_or_add_custom_hook(
            self._recipe_cfg,
            ConfigDict(type="CancelInterfaceHook", init_callback=self.on_hook_initialized),
        )
        if self._time_monitor is not None:
            update_or_add_custom_hook(
                self._recipe_cfg,
                ConfigDict(
                    type="OTXProgressHook",
                    time_monitor=self._time_monitor,
                    verbose=True,
                    priority=71,
                ),
            )
        self._recipe_cfg.log_config.hooks.append({"type": "OTXLoggerHook", "curves": self._learning_curves})

        # make sure model to be in a training mode even after model is evaluated (mmcv bug)
        update_or_add_custom_hook(
            self._recipe_cfg,
            ConfigDict(type="ForceTrainModeHook", priority="LOWEST"),
        )

        # if num_workers is 0, persistent_workers must be False
        data_cfg = self._recipe_cfg.data
        for subset in ["train", "val", "test", "unlabeled"]:
            if subset not in data_cfg:
                continue
            dataloader_cfg = data_cfg.get(f"{subset}_dataloader", ConfigDict())
            workers_per_gpu = dataloader_cfg.get(
                "workers_per_gpu",
                data_cfg.get("workers_per_gpu", 0),
            )
            if workers_per_gpu == 0:
                dataloader_cfg["persistent_workers"] = False
                data_cfg[f"{subset}_dataloader"] = dataloader_cfg

        if self._data_cfg is not None:
            align_data_config_with_recipe(self._data_cfg, self._recipe_cfg)

        if export:
            options["deploy_cfg"] = self._init_deploy_cfg()
            if options.get("precision", None) is None:
                assert len(self._precision) == 1
                options["precision"] = str(self._precision[0])

            options["deploy_cfg"]["dump_features"] = options["dump_features"]
            if options["dump_features"]:
                output_names = options["deploy_cfg"]["ir_config"]["output_names"]
                if "feature_vector" not in output_names and "saliency_map" not in output_names:
                    options["deploy_cfg"]["ir_config"]["output_names"] += ["feature_vector", "saliency_map"]

        self._initialize_post_hook(options)

        logger.info("initialized.")

    def _initialize_post_hook(self, options=None):
        pass

    @abc.abstractmethod
    def _init_recipe(self):
        """Initialize the MPA's target recipe (inclusive of stage type)."""
        raise NotImplementedError("this method should be implemented")

    def _init_model_cfg(self) -> Union[Config, None]:
        """Initialize model_cfg for override recipe's model configuration."""
        raise NotImplementedError("this method should be implemented")

    def _init_train_data_cfg(self, dataset: DatasetEntity) -> Union[Config, None]:
        """Initialize data_cfg for override recipe's data configuration."""
        return ConfigDict(data=dataset) if dataset else self._data_cfg

    def _init_test_data_cfg(self, dataset: DatasetEntity) -> Union[Config, None]:
        """Initialize data_cfg for override recipe's data configuration."""
        return ConfigDict(data=dataset) if dataset else self._data_cfg

    def _init_recipe_hparam(self) -> dict:
        """Initialize recipe hyperparamter as dict."""
        assert self._recipe_cfg is not None

        params = self._hyperparams.learning_parameters
        warmup_iters = int(params.learning_rate_warmup_iters)
        lr_config = (
            ConfigDict(warmup_iters=warmup_iters)
            if warmup_iters > 0
            else ConfigDict(warmup_iters=warmup_iters, warmup=None)
        )

        if params.enable_early_stopping and self._recipe_cfg.get("evaluation", None):
            early_stop = ConfigDict(
                start=int(params.early_stop_start),
                patience=int(params.early_stop_patience),
                iteration_patience=int(params.early_stop_iteration_patience),
            )
        else:
            early_stop = False

        runner = ConfigDict(max_epochs=int(params.num_iters))
        if self._recipe_cfg.get("runner", None) and self._recipe_cfg.runner.get("type").startswith("IterBasedRunner"):
            runner = ConfigDict(max_iters=int(params.num_iters))

        return ConfigDict(
            optimizer=ConfigDict(lr=params.learning_rate),
            lr_config=lr_config,
            early_stop=early_stop,
            data=ConfigDict(
                samples_per_gpu=int(params.batch_size),
                workers_per_gpu=int(params.num_workers),
            ),
            runner=runner,
        )

    def _update_stage_module(self, stage_module: str):
        return stage_module

    def _init_deploy_cfg(self) -> Union[Config, None]:
        base_dir = os.path.abspath(os.path.dirname(self.template_file_path))
        deploy_cfg_path = os.path.join(base_dir, "deployment.py")
        deploy_cfg = None
        if os.path.exists(deploy_cfg_path):
            deploy_cfg = MPAConfig.fromfile(deploy_cfg_path)

            def patch_input_preprocessing(deploy_cfg):

                normalize_cfg = get_configs_by_pairs(
                    self._recipe_cfg.data.test.pipeline,
                    dict(type="Normalize"),
                )
                assert len(normalize_cfg) == 1
                normalize_cfg = normalize_cfg[0]

                options = dict(flags=[], args={})
                # NOTE: OTX loads image in RGB format
                # so that `to_rgb=True` means a format change to BGR instead.
                # Conventionally, OpenVINO IR expects a image in BGR format
                # but OpenVINO IR under OTX assumes a image in RGB format.
                #
                # `to_rgb=True` -> a model was trained with images in BGR format
                #                  and a OpenVINO IR needs to reverse input format from RGB to BGR
                # `to_rgb=False` -> a model was trained with images in RGB format
                #                   and a OpenVINO IR does not need to do a reverse
                if normalize_cfg.get("to_rgb", False):
                    options["flags"] += ["--reverse_input_channels"]
                # value must be a list not a tuple
                if normalize_cfg.get("mean", None) is not None:
                    options["args"]["--mean_values"] = list(normalize_cfg.get("mean"))
                if normalize_cfg.get("std", None) is not None:
                    options["args"]["--scale_values"] = list(normalize_cfg.get("std"))

                # fill default
                backend_config = deploy_cfg.backend_config
                if backend_config.get("mo_options") is None:
                    backend_config.mo_options = ConfigDict()
                mo_options = backend_config.mo_options
                if mo_options.get("args") is None:
                    mo_options.args = ConfigDict()
                if mo_options.get("flags") is None:
                    mo_options.flags = []

                # already defiend options have higher priority
                options["args"].update(mo_options.args)
                mo_options.args = ConfigDict(options["args"])
                # make sure no duplicates
                mo_options.flags.extend(options["flags"])
                mo_options.flags = list(set(mo_options.flags))

            def patch_input_shape(deploy_cfg):
                resize_cfg = get_configs_by_pairs(
                    self._recipe_cfg.data.test.pipeline,
                    dict(type="Resize"),
                )
                assert len(resize_cfg) == 1
                resize_cfg = resize_cfg[0]
                size = resize_cfg.size
                if isinstance(size, int):
                    size = (size, size)
                assert all(isinstance(i, int) and i > 0 for i in size)
                # default is static shape to prevent an unexpected error
                # when converting to OpenVINO IR
                deploy_cfg.backend_config.model_inputs = [ConfigDict(opt_shapes=ConfigDict(input=[1, 3, *size]))]

            patch_input_preprocessing(deploy_cfg)
            if not deploy_cfg.backend_config.get("model_inputs", []):
                patch_input_shape(deploy_cfg)

        return deploy_cfg

    def _load_model_ckpt(self, model: Optional[ModelEntity]):
        if model and "weights.pth" in model.model_adapters:
            # If a model has been trained and saved for the task already, create empty model and load weights here
            buffer = io.BytesIO(model.get_data("weights.pth"))
            model_data = torch.load(buffer, map_location=torch.device("cpu"))

            # set confidence_threshold as well
            self.confidence_threshold = model_data.get("confidence_threshold", self.confidence_threshold)
            if model_data.get("anchors"):
                self._anchors = model_data["anchors"]

            # Get config
            if model_data.get("config"):
                tiling_parameters = model_data.get("config").get("tiling_parameters")
                if tiling_parameters and tiling_parameters["enable_tiling"]["value"]:
                    logger.info("Load tiling parameters")
                    self._hyperparams.tiling_parameters.enable_tiling = tiling_parameters["enable_tiling"]["value"]
                    self._hyperparams.tiling_parameters.tile_size = tiling_parameters["tile_size"]["value"]
                    self._hyperparams.tiling_parameters.tile_overlap = tiling_parameters["tile_overlap"]["value"]
                    self._hyperparams.tiling_parameters.tile_max_number = tiling_parameters["tile_max_number"]["value"]
            return model_data
        return None

    def _load_model_label_schema(self, model: Optional[ModelEntity]):
        # If a model has been trained and saved for the task already, create empty model and load weights here
        if model and "label_schema.json" in model.model_adapters:
            import json

            buffer = json.loads(model.get_data("label_schema.json").decode("utf-8"))
            model_label_schema = LabelSchemaMapper().backward(buffer)
            return model_label_schema.get_labels(include_empty=False)
        return self._labels

    def _load_resume_info(self, model: Optional[ModelEntity]):
        if model and "resume" in model.model_adapters:
            return model.model_adapters.get("resume", False)
        return False

    @staticmethod
    def _get_confidence_threshold(hyperparams):
        confidence_threshold = 0.3
        if hasattr(hyperparams, "postprocessing") and hasattr(hyperparams.postprocessing, "confidence_threshold"):
            confidence_threshold = hyperparams.postprocessing.confidence_threshold
        return confidence_threshold

    @staticmethod
    def _is_docker():
        """Checks whether the task runs in docker container.

        :return bool: True if task runs in docker
        """
        path = "/proc/self/cgroup"
        is_in_docker = False
        if os.path.isfile(path):
            with open(path, encoding="UTF-8") as f:
                is_in_docker = is_in_docker or any("docker" in line for line in f)
        is_in_docker = is_in_docker or os.path.exists("/.dockerenv")
        return is_in_docker

    def cancel_hook_initialized(self, cancel_interface: CancelInterfaceHook):
        """Initialization of cancel_interface hook."""
        logger.info("cancel hook is initialized")
        self.cancel_interface = cancel_interface
        if self.reserved_cancel and self.cancel_interface:
            self.cancel_interface.cancel()

    def unload(self):
        """Unload the task."""
        if self._work_dir_is_temp:
            self._delete_scratch_space()
        if self._is_docker():
            logger.warning("Got unload request. Unloading models. Throwing Segmentation Fault on purpose")
            import ctypes

            ctypes.string_at(0)
        else:
            logger.warning("Got unload request, but not on Docker. Only clearing CUDA cache")
            torch.cuda.empty_cache()
            logger.warning(
                f"Done unloading. " f"Torch is still occupying {torch.cuda.memory_allocated()} bytes of GPU memory"
            )

    def cleanup(self):
        """Clean up work directory if user specified it."""
        if self._work_dir_is_temp:
            self._delete_scratch_space()

    class OnHookInitialized:
        """OnHookInitialized class."""

        def __init__(self, task_instance):
            self.task_instance = task_instance
            self.__findable = False  # a barrier to block segmentation fault

        def __call__(self, cancel_interface):
            """Function call in OnHookInitialized."""
            if isinstance(self.task_instance, int) and self.__findable:
                import ctypes

                # NOTE: BE AWARE OF SEGMENTATION FAULT
                self.task_instance = ctypes.cast(self.task_instance, ctypes.py_object).value
            self.task_instance.cancel_hook_initialized(cancel_interface)

        def __repr__(self):
            """Function repr in OnHookInitialized."""
            return f"'{__name__}.OnHookInitialized'"

        def __deepcopy__(self, memo):
            """Function deepcopy in OnHookInitialized."""
            cls = self.__class__
            result = cls.__new__(cls)
            memo[id(self)] = result
            result.task_instance = self.task_instance
            result.__findable = True  # pylint: disable=unused-private-member
            return result

        def __reduce__(self):
            """Function reduce in OnHookInitialized."""
            return (self.__class__, (id(self.task_instance),))

    def update_override_configurations(self, config):
        """Update override_configs."""
        logger.info(f"update override config with: {config}")
        config = ConfigDict(**config)
        self.override_configs.update(config)

    def set_early_stopping_hook(self):
        """Update Early-stopping Hook."""
        if "early_stop" in self._recipe_cfg:
            remove_custom_hook(self._recipe_cfg, "EarlyStoppingHook")
            early_stop = self._recipe_cfg.get("early_stop", False)
            if early_stop:
                early_stop_hook = ConfigDict(
                    type="LazyEarlyStoppingHook",
                    start=early_stop.start,
                    patience=early_stop.patience,
                    iteration_patience=early_stop.iteration_patience,
                    interval=1,
                    metric=self._recipe_cfg.early_stop_metric,
                    priority=75,
                )
                update_or_add_custom_hook(self._recipe_cfg, early_stop_hook)
            else:
                remove_custom_hook(self._recipe_cfg, "LazyEarlyStoppingHook")



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from mmdet.models.builder import BACKBONES

from ...mmov_model import MMOVModel


@BACKBONES.register_module()
class MMOVBackbone(MMOVModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, *args, **kwargs):
        outputs = super().forward(*args, **kwargs)
        if not isinstance(outputs, tuple):
            outputs = (outputs,)
        # must return tuple
        return outputs

    def init_weights(self, pretrained=None):
        # TODO
        pass



"""Convert dataset: Public dataset (COCO) --> Multi-label dataset (Datumaro format).

this script contains a lot of hard-coding to create .json file that Datumaro can consume.
"""

import argparse
import json

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import Any, Dict, List

from otx.algorithms.detection.utils.data import CocoDataset

multilabel_ann_format = {
    "info": {},
    "categories": {
        "label": {
            "label_groups": [],
            "labels": [],
            "attributes": [],
        }
    },
    "items": [],
}  # type: Dict[str, Any]


def coco_to_datumaro_multilabel(ann_file_path: str, data_root_dir: str, output: str, test_mode: bool = False):
    """Convert coco dataset to datumaro multi-label format.

    Args:
        ann_file_path (str): The path of annotation file (COCO)
        data_root_dir (str): The path of images folder (COCO)
        output (str): Destination path of converted data (CVAT multi-label format)
        test_mode (bool): Omit filtering irrelevant images during COCO dataset initialization for testing purposes.
    """

    # Prepare COCO dataset to load annotations
    coco_dataset = CocoDataset(
        ann_file=ann_file_path,
        data_root=data_root_dir,
        classes=None,
        test_mode=test_mode,
        with_mask=False,
    )

    # Fill the categories part
    # For the multi-label classification,
    # Datumaro will make label_groups and labels
    overall_classes = coco_dataset.get_classes()  # type: List
    for class_name in overall_classes:
        multilabel_ann_format["categories"]["label"]["label_groups"].append(
            {"name": str(class_name), "group_type": "exclusive", "labels": [str(class_name)]}
        )

        multilabel_ann_format["categories"]["label"]["labels"].append(
            {"name": class_name, "parent": "", "attributes": []}
        )

    # Fill the items part
    for item in coco_dataset:
        filename = item["img_info"]["filename"]
        file_id = filename.split(".")[0]
        labels = item["gt_labels"]

        annotations = []
        for i, label in enumerate(labels):
            annotations.append({"id": int(i), "type": "label", "group": 0, "label_id": int(label)})

        multilabel_ann_format["items"].append(
            {"id": str(file_id), "annotations": annotations, "image": {"path": str(filename)}}
        )
    print(f"Saving logfile to: {output}")
    with open(output, "w", encoding="utf-8") as out_file:
        json.dump(multilabel_ann_format, out_file)


def parse_args():
    """Parses command line arguments."""
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument("--ann_file_path", required=True, type=str)
    parser.add_argument("--data_root_dir", required=True, type=str)
    parser.add_argument("--output", required=True, type=str)
    parser.add_argument("--data_format", type=str, default="coco")
    return parser.parse_args()


def main():
    """Main function."""
    args = parse_args()
    if args.data_format == "coco":
        coco_to_datumaro_multilabel(args.ann_file_path, args.data_root_dir, args.output)
    else:
        raise ValueError(f"Unknown data format {args.data_format}.This script only support `coco`.")


if __name__ == "__main__":
    main()



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Optional, Tuple

import numpy as np
import torch

from otx.mpa.utils.logger import get_logger

from ..utils import get_op_name
from .builder import OPS
from .op import Attribute, Operation
from .type_conversions import ConvertV0
from .utils import get_dynamic_shape

logger = get_logger()


NODE_TYPES_WITH_WEIGHT = set(
    [
        "Convolution",
        "GroupConvolution",
        "MatMul",
        "BatchNormInference",
        "Multiply",
        "Divide",
        "Add",
        "Subtract",
    ]
)


@dataclass
class ParameterV0Attribute(Attribute):
    element_type: Optional[str] = field(default=None)

    layout: Optional[Tuple[str]] = field(default=None)
    permute: Optional[Tuple[int]] = field(default=None)
    verify_shape: bool = field(default=True)

    def __post_init__(self):
        super().__post_init__()
        # fmt: off
        valid_element_type = [
            None,
            "u1", "u4", "u8", "u16", "u32", "u64",
            "i4", "i8", "i16", "i32", "i64", "f16", "f32", "boolean", "bf16"
        ]
        # fmt: on
        if self.element_type not in valid_element_type:
            raise ValueError(f"Invalid element_type {self.element_type}. " f"It must be one of {valid_element_type}.")


@OPS.register()
class ParameterV0(Operation[ParameterV0Attribute]):
    TYPE = "Parameter"
    VERSION = 0
    ATTRIBUTE_FACTORY = ParameterV0Attribute

    def forward(self, input):
        # TODO: validate shape
        # need to handle new generated op from reshaped model
        if self.attrs.verify_shape:
            assert self.shape is not None
            ov_shape = self.shape[0]
            torch_shape = list(input.shape)
            for ov_shape_, torch_shape_ in zip(ov_shape, torch_shape):
                if ov_shape_ == -1:
                    continue
                assert ov_shape_ == torch_shape_, f"input shape {torch_shape} does not match with ov shape {ov_shape}"

        if self.attrs.permute:
            input = input.permute(self.attrs.permute)

        return input

    @classmethod
    def from_ov(cls, ov_op):
        op_type = ov_op.get_type_name()
        op_version = ov_op.get_version()
        op_name = get_op_name(ov_op)
        assert cls.TYPE != "" and cls.VERSION >= 0
        assert op_type == cls.TYPE
        assert op_version == cls.VERSION

        attrs = ov_op.get_attributes()
        if "shape" not in attrs:
            shapes = []
            for output in ov_op.outputs():
                shapes.append(get_dynamic_shape(output))
            shapes = tuple(tuple(shape) for shape in shapes)
            attrs["shape"] = shapes

        layout = ov_op.get_layout()
        if not layout.empty:
            layout = layout.to_string()[1:-1].split(",")
            attrs["layout"] = tuple(layout)

            #  N, C, H, W
            input_layout = OrderedDict(
                {
                    "N": 0,
                    "C": 1,
                    "H": 2,
                    "W": 3,
                }
            )
            if not set(layout).symmetric_difference(input_layout.keys()):
                permute = []
                for layout_ in layout:
                    #  N, H, W, C
                    permute.append(input_layout[layout_])
                attrs["permute"] = tuple(permute)

            # TODO: here, we force the batch dim to be dynamic
            # but this should be done when loading ov model
            i = layout.index("N")
            new_shape = []
            for shape in attrs["shape"]:
                new_shape.append([-1 if j == i else k for j, k in enumerate(shape)])
            new_shape = tuple(tuple(shape) for shape in new_shape)
            attrs["shape"] = new_shape

            # change shape and layout based on permute
            if "permute" in attrs and attrs["permute"] != (0, 1, 2, 3):
                assert len(attrs["shape"]) == 1
                permute = []
                for layout_ in input_layout.keys():
                    permute.append(layout.index(layout_))
                new_shape = []
                for shape in attrs["shape"]:
                    new_shape.append([shape[i] for i in permute])
                attrs["shape"] = tuple(tuple(shape) for shape in new_shape)
                attrs["layout"] = tuple([attrs["layout"][i] for i in permute])

        return cls(name=op_name, **attrs)


@dataclass
class ResultV0Attribute(Attribute):
    pass


@OPS.register()
class ResultV0(Operation[ResultV0Attribute]):
    TYPE = "Result"
    VERSION = 0
    ATTRIBUTE_FACTORY = ResultV0Attribute

    def forward(self, input):
        return input


@dataclass
class ConstantV0Attribute(Attribute):
    element_type: str
    offset: int = field(default=0)
    size: int = field(default=0)

    is_parameter: bool = field(default=False)

    def __post_init__(self):
        super().__post_init__()
        # fmt: off
        valid_element_type = [
            "u1", "u4", "u8", "u16", "u32", "u64",
            "i4", "i8", "i16", "i32", "i64", "f16", "f32", "boolean", "bf16"
        ]
        # fmt: on
        if self.element_type not in valid_element_type:
            raise ValueError(f"Invalid element_type {self.element_type}. " f"It must be one of {valid_element_type}.")


@OPS.register()
class ConstantV0(Operation[ConstantV0Attribute]):
    TYPE = "Constant"
    VERSION = 0
    ATTRIBUTE_FACTORY = ConstantV0Attribute

    def __init__(self, *args, **kwargs):
        data = kwargs.pop("data", None)
        if data is None:
            raise KeyError("data is not provided")
        assert isinstance(data, torch.Tensor)
        kwargs["element_type"] = ConvertV0.convert_torch_type(data.dtype)
        super().__init__(*args, **kwargs)
        if self.attrs.is_parameter:
            self.data = torch.nn.Parameter(data)
        else:
            self.register_buffer("data", data)

    def forward(self):
        return self.data

    @classmethod
    def from_ov(cls, ov_op):
        op_type = ov_op.get_type_name()
        op_version = ov_op.get_version()
        op_name = get_op_name(ov_op)
        assert cls.TYPE != "" and cls.VERSION >= 0
        assert op_type == cls.TYPE
        assert op_version == cls.VERSION

        attrs = ov_op.get_attributes()
        attrs["shape"] = tuple(attrs["shape"])

        data = ov_op.get_data()
        if data.dtype == np.uint64:
            data_ = data.astype(np.int64)
            if not np.array_equal(data, data_):
                logger.warning(f"Overflow detected in {op_name}")
            data = torch.from_numpy(data_)
        else:
            data = torch.from_numpy(data)

        in_port_indices = []
        op_node_types = []
        for out_port in ov_op.outputs():
            for in_port in list(out_port.get_target_inputs()):
                in_port_index = in_port.get_index()
                in_port_indices.append(in_port_index)
                node = in_port.get_node()
                op_node_types.append(node.get_type_name())

        # FIXME: need a better way to distinghish if it is parameter or no
        is_parameter = False
        if (
            set(op_node_types).intersection(NODE_TYPES_WITH_WEIGHT)
            and len(in_port_indices) == 1
            and in_port_indices[0] != 0
            and data.numel() > 1
            and (data.is_floating_point() or data.is_complex())
        ):
            is_parameter = True
        attrs["is_parameter"] = is_parameter

        return cls(name=op_name, data=data, **attrs)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# Copyright (c) Open-MMLab. All rights reserved.
from pathlib import Path
from typing import Optional

from mmcv.runner.dist_utils import allreduce_params, master_only
from mmcv.runner.hooks.hook import HOOKS, Hook


@HOOKS.register_module()
class CheckpointHookWithValResults(Hook):
    """Save checkpoints periodically.

    Args:
        interval (int): The saving period. If ``by_epoch=True``, interval
            indicates epochs, otherwise it indicates iterations.
            Default: -1, which means "never".
        by_epoch (bool): Saving checkpoints by epoch or by iteration.
            Default: True.
        save_optimizer (bool): Whether to save optimizer state_dict in the
            checkpoint. It is usually used for resuming experiments.
            Default: True.
        out_dir (str, optional): The directory to save checkpoints. If not
            specified, ``runner.work_dir`` will be used by default.
        max_keep_ckpts (int, optional): The maximum checkpoints to keep.
            In some cases we want only the latest few checkpoints and would
            like to delete old ones to save the disk space.
            Default: -1, which means unlimited.
        sync_buffer (bool): Whether to synchronize buffers in different
            gpus. Default: False.
    """

    def __init__(
        self,
        interval=-1,
        by_epoch=True,
        save_optimizer=True,
        out_dir=None,
        max_keep_ckpts=-1,
        sync_buffer=False,
        **kwargs,
    ):
        self.interval = interval
        self.by_epoch = by_epoch
        self.save_optimizer = save_optimizer
        self.out_dir = out_dir
        self.max_keep_ckpts = max_keep_ckpts
        self.args = kwargs
        self.sync_buffer = sync_buffer
        self._best_model_weight: Optional[Path] = None

    def before_run(self, runner):
        if not self.out_dir:
            self.out_dir = runner.work_dir

    def after_train_epoch(self, runner):
        if not self.by_epoch or not self.every_n_epochs(runner, self.interval):
            return

        if self.sync_buffer:
            allreduce_params(runner.model.buffers())
        save_ema_model = hasattr(runner, "save_ema_model") and runner.save_ema_model
        if save_ema_model:
            backup_model = runner.model
            runner.model = runner.ema_model
        if getattr(runner, "save_ckpt", False):
            runner.logger.info(f"Saving best checkpoint at {runner.epoch + 1} epochs")
            self._save_best_checkpoint(runner)
            runner.save_ckpt = False

        self._save_latest_checkpoint(runner)

        if save_ema_model:
            runner.model = backup_model
            runner.save_ema_model = False

    @master_only
    def _save_best_checkpoint(self, runner):
        """Save the current checkpoint and delete unwanted checkpoint."""
        if self._best_model_weight is not None:  # remove previous best model weight
            prev_model_weight = self.out_dir / self._best_model_weight
            if prev_model_weight.exists():
                prev_model_weight.unlink()

        if self.by_epoch:
            weight_name = f"best_epoch_{runner.epoch + 1}.pth"
        else:
            weight_name = f"best_iter_{runner.iter + 1}.pth"
        runner.save_checkpoint(self.out_dir, filename_tmpl=weight_name, save_optimizer=self.save_optimizer, **self.args)

        self._best_model_weight = Path(weight_name)
        if runner.meta is not None:
            runner.meta.setdefault("hook_msgs", dict())
            runner.meta["hook_msgs"]["best_ckpt"] = str(self.out_dir / self._best_model_weight)

    @master_only
    def _save_latest_checkpoint(self, runner):
        """Save the current checkpoint and delete unwanted checkpoint."""
        if self.by_epoch:
            weight_name_format = "epoch_{}.pth"
            cur_step = runner.epoch + 1
        else:
            weight_name_format = "iter_{}.pth"
            cur_step = runner.iter + 1

        runner.save_checkpoint(
            self.out_dir,
            filename_tmpl=weight_name_format.format(cur_step),
            save_optimizer=self.save_optimizer,
            **self.args,
        )

        # remove other checkpoints
        if self.max_keep_ckpts > 0:
            for _step in range(cur_step - self.max_keep_ckpts * self.interval, 0, -self.interval):
                ckpt_path = self.out_dir / Path(weight_name_format.format(_step))
                if ckpt_path.exists():
                    ckpt_path.unlink()

        if runner.meta is not None:
            cur_ckpt_filename = Path(self.args.get("filename_tmpl", weight_name_format.format(cur_step)))
            runner.meta.setdefault("hook_msgs", dict())
            runner.meta["hook_msgs"]["last_ckpt"] = str(self.out_dir / cur_ckpt_filename)

    def after_train_iter(self, runner):
        if self.by_epoch or not self.every_n_iters(runner, self.interval):
            return

        if hasattr(runner, "save_ckpt"):
            if runner.save_ckpt:
                runner.logger.info(f"Saving checkpoint at {runner.iter + 1} iterations")
                if self.sync_buffer:
                    allreduce_params(runner.model.buffers())
                self._save_checkpoint(runner)
            runner.save_ckpt = False



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import os.path as osp
import random
import time
from typing import Optional

import cv2 as cv
import numpy as np
import pytest
from bson import ObjectId

from otx.algorithms.classification.tasks import (
    ClassificationInferenceTask,
    ClassificationTrainTask,
)
from otx.algorithms.common.tasks.training_base import BaseTask
from otx.api.configuration.helper import create
from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.id import ID
from otx.api.entities.image import Image
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.label_schema import LabelGroup, LabelGroupType, LabelSchemaEntity
from otx.api.entities.metrics import Performance
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import parse_model_template
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.train_parameters import TrainParameters
from otx.api.usecases.tasks.interfaces.export_interface import ExportType
from tests.test_suite.e2e_test_system import e2e_pytest_api

DEFAULT_CLS_TEMPLATE_DIR = osp.join("otx/algorithms/classification", "configs", "efficientnet_b0_cls_incr")


def task_eval(task: BaseTask, model: ModelEntity, dataset: DatasetEntity) -> Performance:
    start_time = time.time()
    result_dataset = task.infer(dataset.with_empty_annotations())
    end_time = time.time()
    print(f"{len(dataset)} analysed in {end_time - start_time} seconds")
    result_set = ResultSetEntity(model=model, ground_truth_dataset=dataset, prediction_dataset=result_dataset)
    task.evaluate(result_set)
    assert result_set.performance is not None
    return result_set.performance


class ClassificationTaskAPIBase:
    @staticmethod
    def generate_label_schema(not_empty_labels, multilabel=False, hierarchical=False):
        assert len(not_empty_labels) > 1

        label_schema = LabelSchemaEntity()
        if multilabel:
            emptylabel = LabelEntity(name="Empty label", is_empty=True, domain=Domain.CLASSIFICATION)
            empty_group = LabelGroup(name="empty", labels=[emptylabel], group_type=LabelGroupType.EMPTY_LABEL)
            for label in not_empty_labels:
                label_schema.add_group(
                    LabelGroup(
                        name=label.name,
                        labels=[label],
                        group_type=LabelGroupType.EXCLUSIVE,
                    )
                )
            label_schema.add_group(empty_group)
        elif hierarchical:
            single_label_classes = ["b", "g", "r"]
            multi_label_classes = ["w", "p"]
            emptylabel = LabelEntity(name="Empty label", is_empty=True, domain=Domain.CLASSIFICATION)
            empty_group = LabelGroup(name="empty", labels=[emptylabel], group_type=LabelGroupType.EMPTY_LABEL)
            single_labels = []
            for label in not_empty_labels:
                if label.name in multi_label_classes:
                    label_schema.add_group(
                        LabelGroup(
                            name=label.name,
                            labels=[label],
                            group_type=LabelGroupType.EXCLUSIVE,
                        )
                    )
                    if empty_group not in label_schema.get_groups(include_empty=True):
                        label_schema.add_group(empty_group)
                elif label.name in single_label_classes:
                    single_labels.append(label)
            if single_labels:
                single_label_group = LabelGroup(
                    name="labels",
                    labels=single_labels,
                    group_type=LabelGroupType.EXCLUSIVE,
                )
                label_schema.add_group(single_label_group)
        else:
            main_group = LabelGroup(
                name="labels",
                labels=not_empty_labels,
                group_type=LabelGroupType.EXCLUSIVE,
            )
            label_schema.add_group(main_group)
        return label_schema

    @staticmethod
    def setup_configurable_parameters(template_dir, num_iters=10):
        model_template = parse_model_template(osp.join(template_dir, "template.yaml"))
        hyper_parameters = create(model_template.hyper_parameters.data)
        hyper_parameters.learning_parameters.num_iters = num_iters
        return hyper_parameters, model_template

    @staticmethod
    def init_environment(params, model_template, multilabel, hierarchical, number_of_images=10):
        resolution = (224, 224)
        if hierarchical:
            colors = [(0, 255, 0), (0, 0, 255), (255, 0, 0), (0, 0, 0), (230, 230, 250)]
            cls_names = ["b", "g", "r", "w", "p"]
            texts = ["Blue", "Green", "Red", "White", "Purple"]
        else:
            colors = [(0, 255, 0), (0, 0, 255)]
            cls_names = ["b", "g"]
            texts = ["Blue", "Green"]
        env_labels = [
            LabelEntity(name=name, domain=Domain.CLASSIFICATION, is_empty=False, id=ID(i))
            for i, name in enumerate(cls_names)
        ]

        items = []

        for _ in range(0, number_of_images):
            for j, lbl in enumerate(env_labels):
                class_img = np.zeros((*resolution, 3), dtype=np.uint8)
                class_img[:] = colors[j]
                class_img = cv.putText(
                    class_img,
                    texts[j],
                    (50, 50),
                    cv.FONT_HERSHEY_SIMPLEX,
                    0.8 + j * 0.2,
                    colors[j - 1],
                    2,
                    cv.LINE_AA,
                )

                image = Image(data=class_img)
                labels = [ScoredLabel(label=lbl, probability=1.0)]
                shapes = [Annotation(Rectangle.generate_full_box(), labels)]
                annotation_scene = AnnotationSceneEntity(kind=AnnotationSceneKind.ANNOTATION, annotations=shapes)
                items.append(DatasetItemEntity(media=image, annotation_scene=annotation_scene))

        rng = random.Random()
        rng.seed(100)
        rng.shuffle(items)
        for i, _ in enumerate(items):
            subset_region = i / number_of_images
            if subset_region >= 0.9:
                subset = Subset.TESTING
            elif subset_region >= 0.6:
                subset = Subset.VALIDATION
            else:
                subset = Subset.TRAINING
            items[i].subset = subset

        dataset = DatasetEntity(items)
        labels_schema = ClassificationTaskAPIBase.generate_label_schema(
            dataset.get_labels(), multilabel=multilabel, hierarchical=hierarchical
        )
        environment = TaskEnvironment(
            model=None,
            hyper_parameters=params,
            label_schema=labels_schema,
            model_template=model_template,
        )
        return environment, dataset


class TestClassificationTaskAPI(ClassificationTaskAPIBase):
    @e2e_pytest_api
    def test_reading_classification_cls_incr_model_template(self):
        classification_template = [
            "efficientnet_b0_cls_incr",
            "efficientnet_v2_s_cls_incr",
            "mobilenet_v3_large_1_cls_incr",
        ]
        for model_template in classification_template:
            parse_model_template(osp.join("otx/algorithms/classification", "configs", model_template, "template.yaml"))

    @e2e_pytest_api
    @pytest.mark.parametrize(
        "multilabel,hierarchical",
        [(False, False), (True, False), (False, True)],
        ids=["multiclass", "multilabel", "hierarchical"],
    )
    def test_training_progress_tracking(self, multilabel, hierarchical):
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_CLS_TEMPLATE_DIR, num_iters=10)
        task_environment, dataset = self.init_environment(
            hyper_parameters, model_template, multilabel, hierarchical, 20
        )
        task = ClassificationTrainTask(task_environment=task_environment)
        print("Task initialized, model training starts.")

        training_progress_curve = []

        def progress_callback(progress: float, score: Optional[float] = None):
            training_progress_curve.append(progress)

        train_parameters = TrainParameters()
        train_parameters.update_progress = progress_callback
        output_model = ModelEntity(
            dataset,
            task_environment.get_model_configuration(),
        )
        task.train(dataset, output_model, train_parameters)

        assert len(training_progress_curve) > 0
        assert np.all(training_progress_curve[1:] >= training_progress_curve[:-1])

    @e2e_pytest_api
    @pytest.mark.parametrize(
        "multilabel,hierarchical",
        [(False, False), (True, False), (False, True)],
        ids=["multiclass", "multilabel", "hierarchical"],
    )
    def test_inference_progress_tracking(self, multilabel, hierarchical):
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_CLS_TEMPLATE_DIR, num_iters=5)
        task_environment, dataset = self.init_environment(
            hyper_parameters, model_template, multilabel, hierarchical, 20
        )
        task = ClassificationInferenceTask(task_environment=task_environment)
        print("Task initialized, model inference starts.")

        inference_progress_curve = []

        def progress_callback(progress: int):
            inference_progress_curve.append(progress)

        inference_parameters = InferenceParameters()
        inference_parameters.update_progress = progress_callback
        task.infer(dataset.with_empty_annotations(), inference_parameters)

        assert len(inference_progress_curve) > 0
        assert np.all(inference_progress_curve[1:] >= inference_progress_curve[:-1])

    @e2e_pytest_api
    @pytest.mark.parametrize(
        "multilabel,hierarchical",
        [(False, False), (True, False), (False, True)],
        ids=["multiclass", "multilabel", "hierarchical"],
    )
    def test_inference_task(self, multilabel, hierarchical):
        # Prepare pretrained weights
        hyper_parameters, model_template = self.setup_configurable_parameters(DEFAULT_CLS_TEMPLATE_DIR, num_iters=2)
        classification_environment, dataset = self.init_environment(
            hyper_parameters, model_template, multilabel, hierarchical, 50
        )
        val_dataset = dataset.get_subset(Subset.VALIDATION)

        train_task = ClassificationTrainTask(task_environment=classification_environment)

        training_progress_curve = []

        def progress_callback(progress: float, score: Optional[float] = None):
            training_progress_curve.append(progress)

        train_parameters = TrainParameters()
        train_parameters.update_progress = progress_callback
        trained_model = ModelEntity(
            dataset,
            classification_environment.get_model_configuration(),
        )
        train_task.train(dataset, trained_model, train_parameters)
        performance_after_train = task_eval(train_task, trained_model, val_dataset)

        # Create InferenceTask
        classification_environment.model = trained_model
        inference_task = ClassificationInferenceTask(task_environment=classification_environment)

        performance_after_load = task_eval(inference_task, trained_model, val_dataset)

        assert performance_after_train == performance_after_load

        # Export
        exported_model = ModelEntity(
            dataset,
            classification_environment.get_model_configuration(),
            _id=ObjectId(),
        )
        inference_task.export(ExportType.OPENVINO, exported_model)



"""This module contains the interface class for tasks that can perform training."""


# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import abc

from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.train_parameters import TrainParameters


class ITrainingTask(metaclass=abc.ABCMeta):
    """A base interface class for tasks which can perform training."""

    @abc.abstractmethod
    def save_model(self, output_model: ModelEntity):
        """Save the model currently loaded by the task to `output_model`.

        This method is for instance used to save the pre-trained weights before training
        when the task has been initialised with pre-trained weights rather than an existing model.

        Args:
            output_model (ModelEntity): Output model where the weights should be stored
        """
        raise NotImplementedError

    @abc.abstractmethod
    def train(
        self,
        dataset: DatasetEntity,
        output_model: ModelEntity,
        train_parameters: TrainParameters,
    ):
        """Train a new model using the model currently loaded by the task.

        If training was successful, the new model should be used for subsequent calls (e.g. `optimize` or `infer`).

        The new model weights should be saved in the object `output_model`.

        The task has two choices:

         - Set the output model weights, if the task was able to improve itself (according to own measures)
         - Set the model state as failed if it failed to improve itself (according to own measures)

        Args:
            dataset (DatasetEntity): Dataset containing the training and validation splits to use for training.
            output_model (ModelEntity): Output model where the weights should be stored
            train_parameters (TrainParameters): Training parameters
        """
        raise NotImplementedError

    @abc.abstractmethod
    def cancel_training(self):
        """Cancels the currently running training process.

        If training is not running, do nothing.
        """
        raise NotImplementedError



"""MMDeploy config of YOLOX model for Detection Task."""

_base_ = ["../../base/deployments/base_detection_dynamic.py"]

ir_config = dict(
    output_names=["boxes", "labels"],
)

backend_config = dict(
    model_inputs=[dict(opt_shapes=dict(input=[-1, 3, 416, 416]))],
)
